//This script is to analyze and quantify PBMCs acquired in z-stacks exported from any microscope system
//Last edited 06-June-2022
//WTS, KM, MK
//v1 has single prompt inputs, works but outputs single FOV -> single CSV table. needs to be updated to output single CSV table with all FOV info per Organ Chip
//v2 updated for GUI testing
//v3 includes GUI with Dialog box for ICR Macro experiment and process run info and compiles all individual FOVs to compiled full statistic table
//v4 has removed some of the noise processing steps that are not applicable to the types of images expected from the acquisition
//v5 updated to pass on to ICR team
//v6 fixed slice processing to compile and analyze z-stacks for Bio-Formats Importer
//v7 updating stack processing, isolating PBMC channel, and saving outputs to correct chip and FOV identifiers
//v8 fixed Excel Read/Write to output in a format compatible with Python script, split into two GUIs to fit on-screen
//v9 update GUI for ease-of-use, include option to input already compiled z-stacks, switch default to batch mode
//v10 validation and bug crushing for wider range of users

//Variable nomenclature:
//	g = global var
// 	l = local var to function
//	arr = array
//	int = integer
//  flt = float
//  str = string
//  bin = boolean
//	path = directory

//Note: some variable names may not use this structure because I used blocks of code found on the imageJ documentation site

print("\\Clear") ;
requires("1.52i");

//**********************************	User Input, Image Global Variables	**********************************

g_int_FontSizeType1 = 24;
g_int_FontSizeType2 = 16;
g_int_FontSizeType3 = 12;

l_bin_redo_input= 0;

//**********************************	Dialog Box GUIs for getting image info for processing and 3D analysis	**********************************

do{
	// Intro Dialog
	// Update version and date
	Dialog.create("Emulate Immune Cell Recruitment Analytical Package");
	Dialog.addMessage("Welcome to the Emulate ICR Analytical Package!\n \n" + 
						"Use this script to enable the 3D segmentation and analysis of adherent immune cells.\n \n" +
						"Accepted image formats: \n" + 
						"  1. Individual images with z-planes and channels separated. \n"  +
						"  (NOT accepted: Compiled z-stacks with channels separated) \n"  +
						"\n \n" + 
						"Fill in the parameters in the following windows to input 3D image slices expressing signal from immune cells.\n \n" + 
						"This script will output the following: \n" + 
						"	- The Excel summary file containing cell 3D centroid positions, for input into the Python Immune Cell Recruitment Analysis Script\n" +
						"	- Maps showing segmentation/annotation overlays\n" +
						"	- Text file containing an audit trail ('log') of user inputs and execution steps\n " + 
						"\n \nClick OK to begin!\n \n \n " +
						"v1.0.0 Jan 16, 2023 \n Emulate, Inc.");
	Dialog.show();
	
	// GUI #1
	Dialog.create("Emulate Immune Cell Recruitment Analytical Package - Image Acquisition Info");

	//Project Run Info - Study ID - File Type Extension - Batch Mode
	Dialog.addMessage("\n ICR Project Info", g_int_FontSizeType2);
	////// Add a note for the identifiers
	Dialog.addMessage("Please make sure all identifiers are entered correctly. Example filename: chip01f01s01ch1.tiff");
	Dialog.addString("Project Name", "Project1");
	
	Dialog.addString("Chip ID in file name", "r01c");
	addGUINote("e.g., chip", -34, 285);
	
	Dialog.addString("FOV in file name", "f");
	addGUINote("e.g., f", -34, 285);
	
	Dialog.addString("Plane in file name", "p");
	addGUINote("e.g., s", -34, 285);
	
	Dialog.addString("Channel in file name", "ch");
	addGUINote("e.g., ch", -34, 285);
	
	Dialog.addString("Image File Type Extension", ".tiff");
	////// Change the note for the filename extension from "e.g., .tiff" to "'.tiff' is different from '.tif'. Please double check the extension."
	addGUINote("'.tiff' is different from '.tif'. Please double check the extension.", -34, 285);

	Dialog.addDirectory("Source Directory", "Please select...");
	addGUINote("Folder containing input images", -5, 180);
	Dialog.addDirectory("Output Directory", "Please select...");
	
	////// Change the note for the output directory from "Empty folder where outputs are stored" to "Please use an EMPTY FOLDER or existing files may be overwritten."
	addGUINote("Please use an EMPTY FOLDER or existing files may be overwritten.\nOutput directory should be different from and outside of the \nsource directory.", -5, 180);
	
	var dialog_x=100;
	var dialog_y=10;
	
	Dialog.setLocation(dialog_x,dialog_y);
	Dialog.show();
	
	//Save variables from GUI #1
	g_str_RunName = Dialog.getString();
	g_str_chipID = Dialog.getString();
	g_str_fovID = Dialog.getString();
	g_str_sliceID = Dialog.getString();
	g_str_channelID = Dialog.getString();
	g_str_extension = Dialog.getString();	

	g_str_dirpath = Dialog.getString();
	g_str_savepath = Dialog.getString();
	
	
	// GUI #2
	//ICR Image Info - Channel Number of PBMC stain - Z-slices in image
	Dialog.create("Emulate Immune Cell Recruitment Analytical Package - Image Processing Info");
	Dialog.addMessage("\n ICR Confocal Image Info", g_int_FontSizeType2);
	////// Remove the option of using images with multiple channels and the variable g_bin_multichannel.
	//Dialog.addCheckbox("Select if files contain multiple fluorescent channels", false);
	
	////// Remove the field of "Number of fluorescent channels in acquisition"
	//Dialog.addNumber("Number of fluorescent channels in acquisition", 2);
	////// Change the name of the field from "Channel number for Immune Cells or PBMCs" to "Channel ID to analyze"
	Dialog.addNumber("Channel ID to analyze", 1);
	////// Add a note for channel number
	addGUINote("An image must contain only 1 channel. Different \nchannels must be separated.", -30, 375);
	////// Add name of the channel
	Dialog.addString("Name of channel/stain", "cell_tracker_CAR-T");
	////// Add a note for the name of the channel
	addGUINote("Example: cell_tracker_CAR-T, caspase, or annexin_V. \nUse underscore instead of space", -30, 375);
	
	////// Remove compiled z-stacks as an input option
	//Dialog.addCheckbox("Select if input files are compiled z-stacks", false);
	Dialog.addNumber("Number of planes", 40);
	Dialog.addNumber("Bottom position", 80);
	addGUINote("(microns) - if unknown, input 0", -30, 375);
	
	Dialog.addNumber("Distance between planes", 4);
	addGUINote("(microns)", -30, 375);
	
	Dialog.addMessage("\n ICR Confocal Image Pre-Processing", g_int_FontSizeType2);
	Dialog.addCheckbox("Despeckle", false);
	Dialog.addCheckbox("Sliding Parabola", false);
	Dialog.addNumber("Rolling Ball Radius (pixels): ", 50);
	// Remove the note “Subtract background is built-in within the script.”
	addGUINote("\nAdvanced users can modify the processing parameters above", 0, 25);
	
	Dialog.setLocation(dialog_x,dialog_y);
	Dialog.show();
	
	//Save variables from GUI #2
	////// Remove the option of using images with multiple channels and the variable g_bin_multichannel.
	//g_bin_multichannel = Dialog.getCheckbox();
	////// Remove the field of "Number of fluorescent channels in acquisition" and the variable g_int_channels
	//g_int_channels = Dialog.getNumber();
	g_int_PBMC_channel = Dialog.getNumber();
	////// Add a variable to store name of the channel
	channel_name = Dialog.getString();
	////// Remove compiled z-stacks as an input option
	g_bin_zstack = false;
	//g_bin_zstack = Dialog.getCheckbox();
	g_int_zslices = Dialog.getNumber();
	g_int_zbottom = Dialog.getNumber();
	g_flt_zdistance = Dialog.getNumber();

	g_bin_despeckle = Dialog.getCheckbox();
	g_bin_subtractBack_sliding = Dialog.getCheckbox();
	g_int_subtractBack_radius = Dialog.getNumber();
	
	// GUI #3
	//ICR 3D Object Processing
	Dialog.create("Emulate Immune Cell Recruitment Analytical Package - Object Processing Info");
	
	Dialog.addMessage("\n ICR 3D Immune Cell Processing Parameters", g_int_FontSizeType2);
	Dialog.addNumber("Min intensity object threshold value", 100);
	Dialog.addNumber("Min cell volume size (voxels): ", 150);;
	// Set the “Max cell volume size (voxels)” parameter (g_int_maxSize) to be a very large value: 9999999999
	Dialog.addNumber("Max cell volume size (voxels): ", 9999999999);

	Dialog.addMessage("\n ICR 3D Annotation Map Overlays", g_int_FontSizeType2);
	g_arr_str_outputLabels = newArray("Centroid Maps", "Object Maps", "Surface Maps");
	g_arr_bin_outputDefaults = newArray(true, true, false);
	Dialog.addCheckboxGroup(1, g_arr_str_outputLabels.length , g_arr_str_outputLabels, g_arr_bin_outputDefaults);
	addGUINote("Select maps to visualize object annotation overlays\n ", 0, 10);
	
	Dialog.addCheckbox("Select to open all images independently during script run", false);
	addGUINote("Warning: requires significant memory availability\n", 0, 30);
	
	Dialog.setLocation(dialog_x,dialog_y);
	Dialog.show();

	//Save variables from GUI #3
	g_int_minThresh = Dialog.getNumber();
	g_int_minSize = Dialog.getNumber();
	g_int_maxSize = Dialog.getNumber();

	g_arr_bin_OutputMapChoice = newArray(g_arr_str_outputLabels.length);
	for (i = 0; i < g_arr_str_outputLabels.length; i++) g_arr_bin_OutputMapChoice[i] = Dialog.getCheckbox();
	
	g_bin_debugMode = Dialog.getCheckbox();
	
	//**********************************	print user selections to Log text file for review by user	**********************************

	print("******** Project Run Info ********");
	
	printVar(g_str_RunName, "g_str_RunName");
	printVar(g_str_extension, "g_str_extension");
	printVar(g_str_chipID, "g_str_chipID");
	printVar(g_str_fovID, "g_str_fovID");
	printVar(g_str_sliceID, "g_str_sliceID");
	printVar(g_str_channelID, "g_str_channelID");
	printVar(g_bin_debugMode , "g_bin_debugMode ");
	printVar(g_str_dirpath, "g_str_dirpath");
	printVar(g_str_savepath , "g_str_savepath ");
	
	print("******** ICR Image and Pre-Processing Info ********");
	
	////// Remove the option of using images with multiple channels and the variable g_bin_multichannel.
	//printVar(g_bin_multichannel, "g_bin_multichannel");
	////// Remove the field of "Number of fluorescent channels in acquisition" and the variable g_int_channels
	//printVar(g_int_channels, "g_int_channels");
	printVar(g_int_PBMC_channel, "g_int_PBMC_channel");
	printVar(g_bin_zstack, "g_bin_zstack");
	printVar(g_int_zslices, "g_int_zslices");
	printVar(g_int_zbottom, "g_int_zbottom");
	printVar(g_flt_zdistance, "g_flt_zdistance");
	
	printVar(g_bin_despeckle,"g_bin_despeckle");
	//printVar(g_bin_subtractBack, "g_bin_subtractBack");
	printVar(g_int_subtractBack_radius, "g_int_subtractBack_radius");
	printVar(g_bin_subtractBack_sliding, "g_bin_subtractBack_sliding");
	
	print("******** ICR 3D Object Processing ********");

	printVar(g_int_minThresh, "g_int_minThresh");
	printVar(g_int_minSize, "g_int_minSize");
	printVar(g_int_maxSize, "g_int_maxSize");

	for (i = 0; i < g_arr_str_outputLabels.length; i++){
		printVar(g_arr_str_outputLabels[i], "g_arr_str_outputLabels["+i+"]");
		printVar(g_arr_bin_OutputMapChoice[i], "g_arr_bin_OutputMapChoice["+i+"]");		
	}
	
	g_str_chipIdentifier = g_str_chipID;
	g_str_fovIdentifier = g_str_fovID;
	g_str_sliceIdentifier = g_str_sliceID;
	g_str_channelIdentifier = g_str_channelID;
	
	g_arr_str_identifiers = newArray(g_str_chipIdentifier, g_str_fovIdentifier, g_str_sliceIdentifier, g_str_channelIdentifier);
	g_arr_str_identifierNames = newArray("Chip Identifier", "FOV Identifier", "Plane Identifier", "Channel Identifier");
	printVar(g_str_chipIdentifier, "g_str_chipIdentifier");
	printVar(g_str_fovIdentifier, "g_str_fovIdentifier");
	printVar(g_str_sliceIdentifier, "g_str_sliceIdentifier");
	printVar(g_str_channelIdentifier, "g_str_channelIdentifier");
	
	selectWindow("Log");	
	l_bin_redo_input = getBoolean("Does the information in the log look correct? \n\n Click No to re-enter info");
	
	// Check if unique identifiers are non-unique
	g_arr_str_identifiers1 = newArray(g_str_chipIdentifier + "{}", g_str_fovIdentifier + "{}", g_str_sliceIdentifier + "{}", g_str_channelIdentifier + "{}");
	for (i = 0; i < g_arr_str_identifiers1.length; i++) {
		for (j = 0; j < g_arr_str_identifiers1.length; j++) {
			if (indexOf(g_arr_str_identifiers1[i], g_arr_str_identifiers1[j]) >= 0 && i != j) {
				l_str_message = "" + g_arr_str_identifierNames[j] + " '" + g_arr_str_identifiers[j] + "' appears in " + g_arr_str_identifierNames[i] + " '" + g_arr_str_identifiers[i] + "'. This script will not be able to analyze this filenaming convention. \n \n Please rename files so that each parameter identifier is unique (i.e., one substring does not appear within another). \n  \n Click OK to re-enter info, or Cancel to exit the macro.";
				Dialog.create("Error - Improper Naming Convention");
				Dialog.addMessage(l_str_message);
				Dialog.show();
				l_bin_redo_input = 0;
			}	
		}
	}
	
} while (!l_bin_redo_input);

print("outside of GUI");

////// Before the program starts, check if the output directory is empty. If not, display a dialog of error message and exit.
files_savepath = getFileList(g_str_savepath);
if (files_savepath.length != 0) {
	exit("Output directory not empty. Please choose an empty one \nto avoid overwriting existing files. Program exiting...");
}

print("start main");
//makes sub output folders
g_arr_str_outputPaths = newArray(0);

print("make folders");
printVar(g_str_RunName, "g_str_RunName");

for (i = 0; i < g_arr_str_outputLabels.length ; i++) {
	printVar(i,"i");
	printVar(g_arr_bin_OutputMapChoice[i], "g_arr_bin_OutputMapChoice["+i+"]");
	printVar(g_str_savepath, "g_str_savepath");
	if (g_arr_bin_OutputMapChoice[i]==1) {
		print("if cond met for i = ", i);	
		////// Remove "Emulate_ICR_AnalyticalTools_" from the filename
		g_arr_str_outputPaths = append(g_arr_str_outputPaths, "" + g_str_savepath + "//" + g_str_RunName + " " + g_arr_str_outputLabels[i]) ;
		printVar(g_arr_str_outputPaths[i], "g_arr_str_outputPaths["+i+"]");
		File.makeDirectory(g_arr_str_outputPaths[i]);
	}
}//end for loop over making all output object map save directories

// Make folder for compiled TIFF stacks
////// Remove "Emulate_ICR_AnalyticalTools_" from the filename
g_str_stack_savepath = g_str_savepath + "TIFFStacks/";
File.makeDirectory(g_str_stack_savepath);

// batch mode by default, unless user chooses debug mode
setBatchMode(!g_bin_debugMode);

// process the images in the input folder
processFolder(g_str_dirpath);
print("processFolder complete");

////// Add a variable to keep track the number of images where no objects are found
img_no_obj = 0;

// Process the stacks through 3D Objects Counter
processStacks();
print("processStacks complete");

updateResults();

// Save Log
selectWindow("Log");
////// Remove "Emulate_ICR_AnalyticalTools_" from the filename
run("Text...", "save=["+ g_str_savepath + "//" + g_str_RunName + "_AuditTrail.txt]");
print(g_str_RunName + "_AuditTrail.txt was saved");
print("Script complete");

////// Add a dialog box to display that the script completed successfully and the number of images where no objects are found
if (img_no_obj == 0) {
	showMessage("Info", "Script completed successfully.");
} else {
	showMessage("Warning", "Script completed successfully. There are " + img_no_obj + " sets of \nimages where no objects are found. Please check \nthresholds and re-run the program before proceeding \nto the next steps.");
}

//**********************************	end Main Process	**********************************

//**********************************	function processFolder	**********************************

// recursive function that will pull list of all files in directory
//look for files that end in user defined extension file type
// and if found send to function processImage
//if folder found it will send to processFolder, works with nested folders

function processFolder (l_str_dirpath) {
	
	printVar(l_str_dirpath, "l_str_dirpath");
	
 	l_arr_str_fnames = getFileList (l_str_dirpath);
 	 
 	for (i = 0; i < l_arr_str_fnames.length; i++) {
 		
    	l_str_fpath = l_str_dirpath + l_arr_str_fnames[i];
    	l_str_filename = l_arr_str_fnames[i];
    	
    	// Checks if this is a file of the correct type
    	l_bin_ifFile = endsWith(l_str_fpath, g_str_extension);
    	
    	l_int_chipIndex = indexOf(l_str_filename, g_str_chipIdentifier)+lengthOf(g_str_chipIdentifier);
		
		l_int_fovIndex = indexOf(l_str_filename, g_str_fovIdentifier) + lengthOf(g_str_fovIdentifier);
    	
    	// Checks if the chip identifier appears in the file name
    	l_bin_ifChip = l_int_chipIndex >= lengthOf(g_str_chipIdentifier);
    	
		if (l_bin_ifFile && l_bin_ifChip){
			l_str_chipNum = substring(l_str_filename, l_int_chipIndex, l_int_chipIndex + 2);
			l_str_fovNum = substring(l_str_filename, l_int_fovIndex, l_int_fovIndex + 2);
			processImage(l_str_fpath,l_str_dirpath, l_str_filename, l_str_chipNum, l_str_fovNum);
		} 
		
		else if (endsWith(l_str_fpath, "/")) {
	   		processFolder(l_str_fpath);
		}
  	}  //end for loop over fnames	
}//end function processFolder

//**********************************	function processImage	**********************************
// function to identify the first plane and PBMC channel
// and save the corresponding slices as one z-stack

function processImage(l_str_temp_fpath, l_str_temp_dirpath, l_str_filename, l_str_chipNum, l_str_fovNum){
	//print("Processing Image: " + l_str_filename);
	// Check if the user is inputting individual slices
	if (!g_bin_zstack) {
		l_bin_first_slice = indexOf(l_str_filename, g_str_sliceIdentifier + "01") >= 0; // check if first slice
		
		if (l_bin_first_slice) {
			// Format input string for BioFormats Importer
			if (g_int_zslices < 10) {
				l_str_sliceInput = g_str_sliceIdentifier + "<01-0" + g_int_zslices + ">"; //ie p<01-09>"
			} else {
				l_str_sliceInput = g_str_sliceIdentifier + "<01-" + g_int_zslices + ">"; //ie "p<01-10>"
			}
			
			l_str_fpathSubstring1 = l_str_temp_dirpath + substring(l_str_filename, 0, indexOf(l_str_filename, g_str_sliceIdentifier+"0"));
			l_str_fpathSubstring2 = substring(l_str_filename, indexOf(l_str_filename, g_str_sliceIdentifier+"0") + lengthOf(g_str_sliceIdentifier) + 2, lengthOf(l_str_filename));
		}
	} else {
		// Import as-is if it's already a stack
		l_str_sliceInput = "";
		l_bin_first_slice = 1;
		
		l_str_fpathSubstring1 = l_str_temp_fpath;
		l_str_fpathSubstring2 = "";
	}
	
	////// Remove the section of "Check if the user is inputting single channel or multichannel images" and only keep the necessary lines
	l_str_PBMC_channel = "" + g_int_PBMC_channel;
	l_bin_PBMC_channel = indexOf(l_str_filename, g_str_channelIdentifier + l_str_PBMC_channel) >= 0;
		
	// Check if the user is inputting single channel or multichannel images
	/*
	if (!g_bin_multichannel && g_int_channels > 1) { //if single channel images, determine if this is the PBMC image
		l_str_PBMC_channel = "" + g_int_PBMC_channel;
		l_bin_PBMC_channel = indexOf(l_str_filename, g_str_channelIdentifier + l_str_PBMC_channel) >= 0;
	} else if (!g_bin_multichannel && g_int_channels == 1) {
		l_str_PBMC_channel = "" + g_int_PBMC_channel;
		l_bin_PBMC_channel = 1;
	} else { // if multichannel, isolate the PBMC channel then continue
		run("Duplicate...", "duplicate channels="+g_int_PBMC_channel + "");
		l_str_PBMC_channel = "" + g_int_PBMC_channel;
		l_bin_PBMC_channel = 1;
	}
	*/
	
	// If first slice and if correct PBMC channel, run Bio-Formats Importer to create stack and save as TIFF
	if (l_bin_first_slice && l_bin_PBMC_channel) {
		
		print("Now being processed: " + l_str_fpath + " Chip " + g_str_chipIdentifier + l_str_chipNum + " FOV " + g_str_fovIdentifier + l_str_fovNum + "");
		
		l_str_stackInput = l_str_fpathSubstring1 + l_str_sliceInput + l_str_fpathSubstring2;

		l_str_bioFormatsInput = "open=["+l_str_temp_fpath+"] autoscale color_mode=Default group_files quiet rois_import=[ROI manager] view=Hyperstack stack_order=XYCZT name=["+l_str_stackInput+"] z_begin=1 z_end="+g_int_zslices+" z_step="+g_flt_zdistance+"";
		printVar(l_str_stackInput, "l_str_stackInput");
		printVar(l_str_bioFormatsInput, "l_str_bioFormatsInput");
		run("Bio-Formats Importer", l_str_bioFormatsInput);
		
		l_str_saveFilepath = g_str_stack_savepath + g_str_chipIdentifier + l_str_chipNum + g_str_fovIdentifier + l_str_fovNum + g_str_channelIdentifier + l_str_PBMC_channel + "_compiled" + g_str_extension;
		
		saveAs("Tiff", l_str_saveFilepath);
		close();
	}
}

//**********************************	function processStacks	**********************************
// function to go through the z-stacks and run 3D objects counter
// then save corresponding outputs

function processStacks () {
	l_arr_str_stackFiles = getFileList(g_str_stack_savepath);
	
	////// Add a dialog box for the number of images found
	if (l_arr_str_stackFiles.length == 0) {
		showMessage("Warning", "No images with the specified filename format are found.");
	} else {
		showMessage("Info", l_arr_str_stackFiles.length + " sets of images with the specified filename format are found.");
	}
	
	for (file = 0; file < l_arr_str_stackFiles.length; file++) {
		l_str_temp_fpath = l_arr_str_stackFiles[file];
		open(g_str_stack_savepath + l_str_temp_fpath);
		
		l_str_name = getTitle;
	  	l_str_name_noext = replace(l_str_name, g_str_extension,"");
	  	
		rename(l_str_name_noext);
		
		//run Noise Pre-processing steps

		if(g_bin_despeckle == 1) run("Despeckle", "stack");
	
		if(g_bin_subtractBack_sliding == 1){
			run("Subtract Background...", "rolling="+g_int_subtractBack_radius+" sliding stack");
		} //else {
		//	run("Subtract Background...", "rolling="+g_int_subtractBack_radius+" stack");
		//}
		// Disable subtract background before object counting. It’s enabled by default even though it’s not selected by the user

		//run 3D objects counter with all outputs from user input choices
	
		//for reference: g_arr_str_outputLabels = newArray("Centroid Maps", "Object Maps", "Surface Maps");
		
		// To Do: Disable “Exclude objects on edges” option
		// l_str_3d_args = "threshold="+g_int_minThresh+" slice=" +g_int_zslices+" min.="+g_int_minSize+" max.="+g_int_maxSize;
		l_str_3d_args = "threshold="+g_int_minThresh+" slice=" +g_int_zslices+" min.="+g_int_minSize+" max.="+g_int_maxSize+" exclude_objects_on_edges ";
		printVar(l_str_3d_args,"l_str_3d_args - pre choice");
		
		if (g_arr_bin_OutputMapChoice[1] == 1) {
			l_str_3d_args = l_str_3d_args + " objects ";
		}
		
		if (g_arr_bin_OutputMapChoice[2] == 1){
			l_str_3d_args = l_str_3d_args + " surfaces ";
		}
		
		if (g_arr_bin_OutputMapChoice[0] == 1){
			l_str_3d_args = l_str_3d_args + " centroids ";
		}
	
		l_str_3d_args = l_str_3d_args + " statistics "+" summary";
		printVar(l_str_3d_args, "l_str_3d_args");
	
		////// Add 3D OC Options to remove labels of object index in output object maps
		run("3D OC Options", "volume surface nb_of_obj._voxels nb_of_surf._voxels integrated_density mean_gray_value std_dev_gray_value median_gray_value minimum_gray_value maximum_gray_value centroid mean_distance_to_surface std_dev_distance_to_surface median_distance_to_surface centre_of_mass bounding_box dots_size=5 font_size=10 store_results_within_a_table_named_after_the_image_(macro_friendly) redirect_to=none");
		run("3D Objects Counter", ""+l_str_3d_args +"");
		
		//save all measured parameters and windows
		//if (nResults > 0) {
			if (g_arr_bin_OutputMapChoice[0] == 1){
				
				l_str_nameCentroidMap = "Centroids map of " + l_str_name_noext ;
				
				//save centroid map
				selectWindow(l_str_nameCentroidMap);
				saveAs("tif", g_arr_str_outputPaths[0] + "//" + l_str_nameCentroidMap +"");
				print(l_str_nameCentroidMap," was saved");	
			}
			
			if (g_arr_bin_OutputMapChoice[1] == 1){
				
				l_str_nameObjectsMap = "Objects map of " + l_str_name_noext ;
				
				//save objects map
				selectWindow(l_str_nameObjectsMap);
				saveAs("tiff",  g_arr_str_outputPaths[1] + "//" + l_str_nameObjectsMap +"");
				print(l_str_nameObjectsMap," was saved");
			}
			
			if (g_arr_bin_OutputMapChoice[2] == 1){
				
				l_str_nameSurfaceMap = "Surface map of " + l_str_name_noext ;	
				
				//save surgace map
				selectWindow(l_str_nameSurfaceMap);
				saveAs("tiff", g_arr_str_outputPaths[2] + "//" + l_str_nameSurfaceMap +"");
				print(l_str_nameSurfaceMap," was saved");
			}
		//}
		
		l_str_nameStatistics = "Statistics for " + l_str_name_noext ;
		
		// Variables for Chip and FOV ID
		l_int_chipIndex = indexOf(l_str_temp_fpath, g_str_chipIdentifier)+lengthOf(g_str_chipIdentifier);
		l_str_chipNum = substring(l_str_temp_fpath, l_int_chipIndex, l_int_chipIndex + 2);
		printVar(l_str_chipNum, "l_str_chipNum");
		
		l_int_fovIndex = indexOf(l_str_temp_fpath, g_str_fovIdentifier) + lengthOf(g_str_fovIdentifier);
		l_str_fovNum = substring(l_str_temp_fpath, l_int_fovIndex, l_int_fovIndex + 2);
		printVar(l_str_fovNum, "l_str_fovNum");

		selectWindow(l_str_nameStatistics);
		Table.rename(l_str_nameStatistics, "Results");
		printVar(nResults,"nResults");
		l_flt_zPositionMicrons = newArray(nResults);
		
		//calculate and add z-position in microns
		for (i = 0; i < nResults(); i++) {
		    l_flt_zPositionMicrons[i] = (g_int_zbottom + ((getResult("Z", i))*g_flt_zdistance));
		}
		
		for (i = 0; i < nResults(); i++) {
			setResult("Chip ID", i, l_str_chipNum);
			setResult("FOV ID", i, l_str_fovNum);
			setResult("Z-pos (micron)",i, l_flt_zPositionMicrons[i]);
		}
		
		updateResults();	
		
		// Excel Read/Write for full statistics
		////// Change the filename of the excel spreadsheet to include channel index and channel name
		////// Remove "Emulate_ICR_AnalyticalTools_" from the filename
		l_str_outputFile = g_str_savepath + "Full_Statistics_ch" + g_int_PBMC_channel + "_" + channel_name + ".xlsx]";
		////// Change the name of the spreadsheet
		////// Now instead of naming the spreadsheet as "Chip_03", the spreadsheet is named as the correct full chip name "r02c03"
		readWriteParameters = "file=[" + l_str_outputFile + " stack_results no_count_column sheet=" + g_str_chipIdentifier + l_str_chipNum + "";
		run("Read and Write Excel", readWriteParameters);
		////// Include the chip identifier in the print statement
		////// The print statement becomes “Chip r02c03 FOV 03 was saved to Full Statistics” instead of “Chip 03 FOV 03 was saved to Full Statistics”
		print("Chip " + g_str_chipIdentifier + l_str_chipNum + " FOV " + l_str_fovNum + " was saved to Full Statistics");
		
		// Excel Read/Write for XYZ statistics for Python script input
		if (nResults > 0) {
			l_arr_str_tableHeadings = newArray("Volume (micron^3)", "Surface (micron^2)", "Nb of obj. voxels",	"Nb of surf. voxels", "IntDen", "Mean", "StdDev", "Median", "Min", "Max", /*"X", "Y",*/ "Z", "Mean dist. to surf. (micron)", "SD dist. to surf. (micron)", "Median dist. to surf. (micron)", "XM", "YM", "ZM", "BX", "BY", "BZ", "B-width", "B-height", "B-depth", "Chip ID", "FOV ID"/*, "Z-pos (micron)"*/);
		} else if (nResults == 0) {
			l_arr_str_tableHeadings = newArray("Volume (micron^3)", "Surface (micron^2)", "Nb of obj. voxels",	"Nb of surf. voxels", "IntDen", "Mean", "StdDev", "Median", "Min", "Max", /*"X", "Y",*/ "Z", "Mean dist. to surf. (micron)", "SD dist. to surf. (micron)", "Median dist. to surf. (micron)", "XM", "YM", "ZM", "BX", "BY", "BZ", "B-width", "B-height", "B-depth"/*, "Chip ID", "FOV ID", "Z-pos (micron)"*/);
		}
		
		////// Increment the variable to keep track the number of images where no objects are found
		if (nResults == 0){
			img_no_obj++;
		}
		
		selectWindow("Results");
		for (i = 0; i < l_arr_str_tableHeadings.length; i++) {
			Table.deleteColumn(l_arr_str_tableHeadings[i]);
		}
		
		if (nResults > 0) Table.renameColumn("Z-pos (micron)", "Z");
		updateResults();
		
		////// Remove the output Emulate_ICR_AnalyticalTools_Position_Statistics.xlsx
		/*
		l_str_outputFile2 = g_str_savepath + "Emulate_ICR_AnalyticalTools_Position_Statistics.xlsx]";
		readWriteParameters2 = "file=[" + l_str_outputFile2 + " stack_results dataset_label=[Chip_"+l_str_chipNum+"] no_count_column sheet=Chip_" + l_str_chipNum + "";
		run("Read and Write Excel", readWriteParameters2);
		print("Chip " + l_str_chipNum + " FOV " + l_str_fovNum + " was saved to Position Statistics");
		*/

		close("Results");
		close("*");
	}
	
}

//function append, inputs arr and appends with value to end of array

function append(arr, value) {
     arr2 = newArray(arr.length+1);
     for (i=0; i<arr.length; i++)
        arr2[i] = arr[i];
     arr2[arr.length] = value;
     return arr2;
}

//function printVar, inputs variable value and variable name as a string and prints to Log, for easy variable tracking and debugging

function printVar(var_value, var_str){
	
	print(var_str + " : " + var_value);
	
}

// function addGuiNote adds a message into the gui
function addGUINote(string, yvalue, xvalue) {
	Dialog.setInsets(yvalue, xvalue, 0);
	Dialog.addMessage(string, g_int_FontSizeType3, "blue");
}
