"""
classification.py

Description: 
    Functions that classify objects according to their
    centroid coordinates relative to the membrane estimates.
    
"""


__author__ = "John Lin"
__contact__ = "john.lin@emulatebio.com"
__date__ = "2023-05-28"
__status__ = "Prototype"


import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from utils import plot_to_png


def fit_poly_membrane(
    subset_df, memb_est_z_lower_bound, memb_est_z_upper_bound
):
    """
    Estimate the shape and location of the membrane using the
    given data frame. It uses a sliding window approach to find
    location in which the window contains the maximum number of 
    data points. Then, a 2nd degree polynomial is fit to the 
    data points in that window to estimate the shape. That
    location and shape corresponds to the bottom end of the 
    membrane. 
    
    From ICR_clustering_v.1.0.0.py

    The core process has been updated.
    
    Parameters
    ----------
    subset_df : numpy.DataFrame
        a subset of the full statistics table produced by the ImageJ
        script. Theoretically, it should contain only 1 unique Chip ID
        and 1 unique FOV ID.
        
    memb_est_z_lower_bound : int
        The lowest point on the z-axis, in micron, where the membrane
        can be possibly found.
        
    memb_est_z_upper_bound : int
        The highest point on the z-axis, in micron, where the membrane
        can be possibly found.
    """
        
    X_IC = np.array(subset_df["X"])
    Y_IC = np.array(subset_df["Y"])
    Z_IC = np.array(subset_df["Z-pos (micron)"])

    #print(X_IC)
    #print(Y_IC)
    #print(Z_IC)
    #print(len(X_IC))
    #print(len(Y_IC))
    #print(len(Z_IC))
    
    ############################################################
    #Filtering on the Z axis with window size 30um
    ############################################################
    win_size = 30
    obj_counts = []

    # Window size must be greater than 0
    assert win_size > 0
    
    # Use the lower bound of the window as the reference
    # The lower bound of the window stops at 
    # memb_est_z_upper_bound - win_size,
    # so the upper bound of the window can stop at exactly 
    # memb_est_z_upper_bound, without going over the
    # user-defined upper bound
    for lower_bound_win in range(
        memb_est_z_lower_bound, 
        memb_est_z_upper_bound - win_size + 1
    ):
        
        # Identify objects with z-coords that fall into this window
        upper_bound_win = lower_bound_win + win_size
        obj_in_win = (Z_IC >= lower_bound_win) & (Z_IC < upper_bound_win)
        
        # Count the number of "True" in the array, which are
        # objects that fall into this window
        num_obj_in_win = np.count_nonzero(obj_in_win)
        #print(num_obj_in_win)
        
        # Keep track of the count for every window
        obj_counts.append(num_obj_in_win)
        
        # Check the upper and lower bounds of the window to make
        # sure they are valid
        assert memb_est_z_lower_bound <= lower_bound_win and lower_bound_win < memb_est_z_upper_bound
        assert memb_est_z_lower_bound < upper_bound_win and upper_bound_win <= memb_est_z_upper_bound
        assert lower_bound_win < upper_bound_win
        
    #print(obj_counts)
    
    # Find the index of the array with the max count
    # The index is the offset from memb_est_z_lower_bound
    index_win_max_count = np.argmax(obj_counts)
    #print(index_win_max_count)
    
    # Add the offset to get the lower bound of the window 
    # with the max count
    lower_bound_win_max_count = memb_est_z_lower_bound + index_win_max_count
    
    # Add the window size to the lower bound of the window to
    # get to the upper bound of the window with the max count
    upper_bound_win_max_count = lower_bound_win_max_count + win_size
    #print(lower_bound_win_max_count)
    #print(upper_bound_win_max_count)
    
    # Identify objects with z-coords that fall into this window
    obj_in_win_max_count = (Z_IC >= lower_bound_win_max_count) & (Z_IC <= upper_bound_win_max_count)
    
    # Extract the x, y, and z-coords of these objects
    X2_IC = X_IC[obj_in_win_max_count]
    Y2_IC = Y_IC[obj_in_win_max_count]
    Z2_IC = Z_IC[obj_in_win_max_count]

    #plt.scatter(X2_IC, Z2_IC)
    #plt.show()


    ########################################################################################################################
    # fit second degreee polynomial to the data in order to estimate the location of the membrane
    ########################################################################################################################
    p = np.polyfit(list(X2_IC), list(Z2_IC), 2)

    pp = np.poly1d(p)
    
    
    return p, pp


def classify_obj(subset_df, chip_id, fov_id, p, pp, output_dir_path):
    """
    Classify data points according to the location of their centroids
    relative to the membrane estimates. The bottom, or lower end of
    the membrane is estimated by fit_poly_membrane(). The top, 
    or the upper end, is calculated as 50 micron above the lower end.
    
    From ICR_clustering_v.1.0.0.py

    The core classification process remains unchanged. Only the
    formats of inputs and outputs are changed.
    
    Parameters
    ----------
    subset_df : numpy.DataFrame
        a subset of the full statistics table produced by the ImageJ
        script.
    
    chip_id : string
        chip_id
        
    fov_id : int
        fov_id in int
    
    p : numpy.ndarray
        An array that contains the coefficient
        
    pp : numpy.poly1d
        A poly1d object that facilitates function evaluation
        given coordinates
        
    output_dir_path : string
        The path to the output directory
    """
    
    
    X_IC = np.array(subset_df["X"])
    Y_IC = np.array(subset_df["Y"])
    Z_IC = np.array(subset_df["Z-pos (micron)"])

    #print(X_IC)
    #print(Y_IC)
    #print(Z_IC)
    #print(len(X_IC))
    #print(len(Y_IC))
    #print(len(Z_IC))
    #print(p)
    #print(pp)


    ########################################################################################################################
    #Perform the PBMCs clustering into 4 categories
    ########################################################################################################################
    count1=0
    count2=0
    count3=0
    count4=0
    MembHight = 50


    class_1_pos, class_2_pos, class_3_pos, class_4_pos = [], [], [], []
    class_1_idx, class_2_idx, class_3_idx, class_4_idx = [], [], [], []


    for i in range(0, len(X_IC)):
        Pol_bot =  p[0]*(X_IC[i])**2 + p[1]*(X_IC[i])+ p[2]
        Pol_Top =  p[0]*(X_IC[i])**2 + p[1]*(X_IC[i])+ p[2]+ MembHight
        #if the point is below the bottom of the membrane and its distance 
        #is larger than 10um then it belongs to category 1
        if((Pol_bot-10 > Z_IC[i])):

            count1 = count1 + 1
            class_1_pos.append([X_IC[i], Y_IC[i], Z_IC[i]])
            class_1_idx.append(i)


        #if the point is below or up from the bottom of the membrane and its distance 
        #is less than 10um then it belongs to category 2    
        elif ((Pol_bot-10 < Z_IC[i]) & (Pol_bot+10 > Z_IC[i])):
            count2 = count2 + 1
            class_2_pos.append([X_IC[i], Y_IC[i], Z_IC[i]])
            class_2_idx.append(i)


        #if the point is above the bottom of the membrane and its distance 
        #is greater than 10um AND is below the top of the membrane and its 
        #distance is greater than 10um then it belongs to category 3 (inside the membrane)    
        elif ((Pol_bot+10 < Z_IC[i]) & (Pol_Top-10 > Z_IC[i])):
            count3 = count3 + 1
            class_3_pos.append([X_IC[i], Y_IC[i], Z_IC[i]])
            class_3_idx.append(i)

        else:
            count4 = count4 + 1
            class_4_pos.append([X_IC[i], Y_IC[i], Z_IC[i]])
            class_4_idx.append(i)

    xp = np.linspace(min(X_IC), max(X_IC), 100)

    # Create a scatter plot to show the classification results
    plot_classification(
        class_1_pos, class_2_pos, class_3_pos, class_4_pos, 
        xp, pp, MembHight, chip_id, fov_id,
        output_dir_path
    )
    
    
    class_pos = [class_1_pos, class_2_pos, class_3_pos, class_4_pos]
    class_idx = [class_1_idx, class_2_idx, class_3_idx, class_4_idx]
    
    return class_pos, class_idx


def plot_classification(class_1_pos, class_2_pos, class_3_pos, class_4_pos, xp, pp, MembHight, chip_id, fov_id, output_dir_path):
    """
    Plot object centroids and membrane estimates in a scatter plot.
    
    From ICR_clustering_v.1.0.0.py. Use new variables class_<n>_pos
    instead of old category<n>_points.

    The object centroids are labeled with different colors to indicate
    their classes. 
    
    Red: on top and above the membrane
    Black: within the membrane
    Green: at the bottom of the membrane
    Blue: below the membrane
    
    Parameters
    ----------
    class_1_pos : list of list
        A 2D list. The rows are the data points in this class. The
        columns are the x, y, and z coordinates, respectively.
        
    class_2_pos : list of list
        A 2D list. The rows are the data points in this class. The
        columns are the x, y, and z coordinates, respectively.
        
    class_3_pos : list of list
        A 2D list. The rows are the data points in this class. The
        columns are the x, y, and z coordinates, respectively.
        
    class_4_pos : list of list
        A 2D list. The rows are the data points in this class. The
        columns are the x, y, and z coordinates, respectively.
        
    xp : numpy.ndarray
        x coordinates of the membrane estimate
        
    pp : numpy.poly1d
        A poly1d object that facilitates function evaluation
        given coordinates
        
    MembHight : int
        The height difference between the bottom and the top
        ends of the membrane
        
    chip_id : string
        chip_id
        
    fov_id : string
        fov_id
                    
    output_dir_path : string
        The path to the output directory
    """
    
    class_1_pos = np.array(class_1_pos)
    class_2_pos = np.array(class_2_pos)
    class_3_pos = np.array(class_3_pos)
    class_4_pos = np.array(class_4_pos)


    plt.plot(xp, pp(xp), '--', xp, pp(xp)+MembHight, '--')

    if(len(class_1_pos) > 0):
        plt.scatter(class_1_pos[:,0], class_1_pos[:,2], marker='o', color='blue')
    if(len(class_2_pos) > 0):
        plt.scatter(class_2_pos[:,0], class_2_pos[:,2], marker='o', color='green')
    if(len(class_3_pos) > 0):
        plt.scatter(class_3_pos[:,0], class_3_pos[:,2], marker='o', color='black')
    if(len(class_4_pos) > 0):
        plt.scatter(class_4_pos[:,0], class_4_pos[:,2], marker='o', color='red')


    plt.plot(xp, pp(xp), '--', xp, pp(xp)+MembHight, '--',color='black')

    plt.xlabel("X-Position ($\mu M$)")
    plt.ylabel("Z-Position ($\mu M$)")
    plt.title(f"{chip_id} fov_{str(fov_id)} object centroid map")

    # Write object map to an image
    output_category = "2D_object_centroid_map"
    file_name = f"{chip_id}_fov_{str(fov_id)}_map.png"
    plot_to_png(output_dir_path, output_category, file_name, plt)
    plt.show()
    
    
class MembraneEstimate:
    """
    A container that stores the coefficients of a second degree
    polynomial which is fit to all data points of all FOV, 
    to estimate the membrane shape and location of the 
    given chip.
    
    Attributes
    ----------
    p : numpy.ndarray
        An array that contains the coefficient
        
    pp : numpy.poly1d
        A poly1d object that facilitates function evaluation
        given coordinates
        
    Methods
    -------
    None
    """
    
    def __init__(self, p, pp):
        self.p = p
        self.pp = pp
        
        
def estimate_memb_all_chips(file_path, min_Z_memb, max_Z_memb, output_dir_path):
    """
    Estimate membrane shape and location using all FOV of each chip,
    for all chips. A wrapper for fit_poly_membrane, which calculates
    the estimate for each chip. The results of all chips are stored
    in a dictionary.
    
    Parameters
    ----------
    file_path : string
        The path to the excel spreadsheet that contains all
        statistics of all objects
        
    min_Z_memb : int
        The lowest point on the z-axis, in micron, where the membrane
        can be possibly found.
        
    max_Z_memb : int
        The highest point on the z-axis, in micron, where the membrane
        can be possibly found.
            
    output_dir_path : string
        The path to the output directory
        
    Returns
    -------
    memb_est_all_chips : dict
        Store the coefficients of membrane estimates using
        MembraneEstimate objects as containers. The keys are chip_id.
    """

    # Load the excel file
    xls = pd.ExcelFile(file_path)

    # Data of different chip ID are stored in different sheets
    sheet_names = xls.sheet_names

    # A dict that stores membrane estimates of all chip ID
    memb_est_all_chips = {}

    # Perform analysis for different chip ID
    for sheet in range(0, len(xls.sheet_names)):

        chip_id = xls.sheet_names[sheet]
        print(chip_id)

        df_chip = pd.read_excel(xls, chip_id, header=1)

        # Estimate the membrane shape and location using all FOV in
        # the same chip. In classificaiton, all FOV in the same chip
        # use the same fit coefficients
        p, pp = fit_poly_membrane(df_chip, min_Z_memb, max_Z_memb)

        print(p)
        print(pp)
        
        fov_id = "all"
        # Call classification but only for the purpose of making
        # a 2D centroid map for the entire chip by aggregating all
        # data points from all FOV.
        _, _ = classify_obj(
                df_chip, chip_id, fov_id, p, pp, output_dir_path
        )

        # Use MembraneEstimate object as a container to store the 
        # coefficients, and assign chip_id to it as the key, then
        # store the container in a dict.
        memb_est_all_chips[chip_id] = MembraneEstimate(p, pp)
        
    return memb_est_all_chips

