"""
colocalization.py

Description: 
    Functions that perform co-localization between two images
    using the fractional overlap method.
    
"""


__author__ = "John Lin"
__contact__ = "john.lin@emulatebio.com"
__date__ = "2023-05-30"
__status__ = "Prototype"


import numpy as np
import pandas as pd
from skimage import color
from skimage.util import img_as_ubyte

from utils import data_frame_to_csv, display_img, write_tiff


def fractional_overlap(obj_map_stain_A, obj_map_stain_B, threshold_ratio_overlap=0.1):
    """
    The core co-localization function using the fractional overlap
    method.
    Overlap = volume intersection / volume of object from stain A.
    
    Use stain A as the baseline/reference. An object from stain A is 
    co-localized if there are objects from stain B that overlap with
    it and the ratio of overlap exceeds the threshold.
    
    Parameters
    ----------
    obj_map_stain_A : numpy.ndarray
        A 3D image of object map in the shape of (N, H, W), where N is
        the number of z-slices. The object map is produced by 3D
        object counter in ImageJ. In the image, the background is 0, 
        and objects are labeled by pixels with values equal to its
        index. All pixel values are unsigned int, which can be 8-bit,
        16-bit, or others, depending on the the output format set
        by ImageJ
        
    obj_map_stain_B : numpy.ndarray
        A 3D image of object map in the shape of (N, H, W), where N is
        the number of z-slices. The object map is produced by 3D
        object counter in ImageJ. In the image, the background is 0, 
        and objects are labeled by pixels with values equal to its
        index. All pixel values are unsigned int, which can be 8-bit,
        16-bit, or others, depending on the the output format set
        by ImageJ
        
    threshold_ratio_overlap : float
        The minimum ratio of overlap to be considered as co-localized
        
    Returns
    -------
    all_coloc_stats : list of dict
        Store the statistics of co-localization for each object.
        Will be converted to pandas data frame.
        
    all_coloc_obj_stain_B : numpy.ndarray
        A 1D array of int that stores the indices of co-localized
        objects from stain B.
    """
    
    # Store stats related to co-localization
    all_coloc_stats = []

    # The indices of co-localized objects from stain B
    all_coloc_obj_stain_B = np.array([], dtype=int)
    
    
    # count the number of objects from stain A
    # Make sure to convert type to int
    n_obj_stain_A = int(np.max(obj_map_stain_A))
    print(f"number of objects in stain A: {n_obj_stain_A}")

    # count the number of objects from stain B
    # Make sure to convert type to int
    n_obj_stain_B = int(np.max(obj_map_stain_B))
    print(f"number of objects in stain B: {n_obj_stain_B}")
    

    # Object index starts from 1 to n_obj_stain_A+1.
    # The background is 0.
    for idx_obj_stain_A in range(1, n_obj_stain_A+1):

        # Store the index of object from stain A
        stats = {"idx_obj_stain_A": idx_obj_stain_A}
    
        # Mask of the stain A object
        obj_mask_stain_A = obj_map_stain_A == idx_obj_stain_A
        
        # Calculate the volume
        volume_obj_stain_A = np.count_nonzero(obj_mask_stain_A)
        
        #print(f"volume_obj_stain_A: {volume_obj_stain_A}")
        stats["volume_obj_stain_A"] = volume_obj_stain_A
        # obj_map_stain_A[obj_mask_stain_A]

        # Apply the stain A object mask to stain B object map 
        obj_mask_A_in_B = obj_map_stain_B[obj_mask_stain_A]
        #obj_mask_A_in_B
        
        # Get the indices of objects from stain B that overlap
        idx_coloc_obj_stain_B = np.unique(obj_mask_A_in_B)
        idx_coloc_obj_stain_B = np.array(idx_coloc_obj_stain_B, dtype=int)

        # Discard the background (pixel value = 0)
        if idx_coloc_obj_stain_B[0] == 0:
            idx_coloc_obj_stain_B = idx_coloc_obj_stain_B[1:]
        else:
            print("complete overlap")

        # Number of objects from stain B that overlap
        #print(f"n_coloc_obj_stain_B: {len(idx_coloc_obj_stain_B)}")
        stats["n_coloc_obj_stain_B"] = len(idx_coloc_obj_stain_B)

        #display(idx_coloc_obj_stain_B)
        stats["idx_coloc_obj_stain_B"] = str(idx_coloc_obj_stain_B)

        # The intersection volume is the number of pixels with 
        # positive value
        volume_intersection = np.count_nonzero(obj_mask_A_in_B)
        #print(f"volume_intersection: {volume_intersection}")
        stats["volume_intersection"] = volume_intersection
        
        # The ratio of overlap
        ratio_overlap = volume_intersection / volume_obj_stain_A
        #print(f"ratio_overlap: {ratio_overlap}")
        stats["ratio_overlap"] = ratio_overlap

        # Co-localization iff the overlap exceeds the threshold
        if ratio_overlap > threshold_ratio_overlap:
            all_coloc_stats.append(stats)
            all_coloc_obj_stain_B = np.append(
                all_coloc_obj_stain_B, idx_coloc_obj_stain_B
            )

        #print("\n")

    return all_coloc_stats, all_coloc_obj_stain_B


def make_stain_coloc_map(obj_map_stain_A, obj_map_stain_B, all_coloc_obj_stain_A, all_coloc_obj_stain_B):
    """
    Create an object map showing co-localized objects from stains A 
    and B and their intersections. Only co-localized objects
    are shown.
    
    Parameters
    ----------
    obj_map_stain_A : numpy.ndarray
        A 3D image of object map in the shape of (N, H, W), where N is
        the number of z-slices. The object map is produced by 3D
        object counter in ImageJ. In the image, the background is 0, 
        and objects are labeled by pixels with values equal to its
        index. All pixel values are unsigned int, which can be 8-bit,
        16-bit, or others, depending on the the output format set
        by ImageJ
        
    obj_map_stain_B : numpy.ndarray
        A 3D image of object map in the shape of (N, H, W), where N is
        the number of z-slices. The object map is produced by 3D
        object counter in ImageJ. In the image, the background is 0, 
        and objects are labeled by pixels with values equal to its
        index. All pixel values are unsigned int, which can be 8-bit,
        16-bit, or others, depending on the the output format set
        by ImageJ
        
    all_coloc_obj_stain_A : numpy.ndarray
        A 1D array of int that stores the indices of co-localized
        objects from stain A.
        
    all_coloc_obj_stain_B : numpy.ndarray
        A 1D array of int that stores the indices of co-localized
        objects from stain B.
        
    Returns
    -------
    stain_coloc_map_rgb : numpy.ndarray, dtype="uint8" 
        A 4D image of object map in the shape of (N, H, W, 3), where
        N is the number of z-slices. It has 3 color channels RGB.
        Objects from stains A are labeled red. Objects from stains 
        B are labeled green. Their intersections are labeled
        yellow.
    """
    
    # Create an empty image
    # dtype of uint8 is acceptable because the max value in the image
    # is 3
    stain_coloc_map = np.full(obj_map_stain_A.shape, 0, dtype="uint8")
    #stain_coloc_map

    # Label all co-localized stain A objects in the co-localization_map
    for idx_obj_stain_A in all_coloc_obj_stain_A:
        #print(idx_obj_stain_A)
        obj_mask_stain_A = obj_map_stain_A == idx_obj_stain_A
        stain_coloc_map[obj_mask_stain_A] = 1

    #display_img(stain_coloc_map[16])

    
    # Label all co-localized stain B objects in the co-localization_map
    for idx_obj_stain_B in all_coloc_obj_stain_B:
        #print(idx_obj_stain_B)
        obj_mask_stain_B = obj_map_stain_B == idx_obj_stain_B
        stain_coloc_map[obj_mask_stain_B] = 2

    #display_img(stain_coloc_map[16])
    
    
    # Create an object mask of all objects from stain B
    obj_mask_all_stain_B = obj_map_stain_B > 0

    # Label all intersections between stain A and stain B objects in 
    # the co-localization map
    for idx_obj_stain_A in all_coloc_obj_stain_A:
        obj_mask_stain_A = obj_map_stain_A == idx_obj_stain_A
        
        # "Logical and" operator is used for intersection
        intersetion = obj_mask_all_stain_B & obj_mask_stain_A
        stain_coloc_map[intersetion] = 3
        
    #display_img(stain_coloc_map[16])
    

    # Convert grayscale to RGB
    stain_coloc_map_rgb = color.label2rgb(
        stain_coloc_map,
        colors=["red", "green", "yellow"],
        bg_label=0
    )
    
    # Convert data type from float32 (ragne 0.0 to 1.0) 
    # to uint8 (range 0 to 255)
    stain_coloc_map_rgb = img_as_ubyte(stain_coloc_map_rgb)
    #display_img(stain_coloc_map_rgb[16])
    
    
    return stain_coloc_map_rgb


def _get_all_non_coloc_obj(obj_map, all_coloc_obj):
    """
    A helper function for make_cell_state_obj_map that gets the indices
    of non-co-localized objects.
    
    Parameters
    ----------
    obj_map : numpy.ndarray
        A 3D image of object map in the shape of (N, H, W), where N is
        the number of z-slices. The object map is produced by 3D
        object counter in ImageJ. In the image, the background is 0, 
        and objects are labeled by pixels with values equal to its
        index. All pixel values are unsigned int, which can be 8-bit,
        16-bit, or others, depending on the the output format set
        by ImageJ
        
    all_coloc_obj : numpy.ndarray
        A 1D array of int that stores the indices of co-localized
        objects.
        
    Returns
    -------
    all_non_coloc_obj : numpy.ndarray
        A 1D array of int that stores the indices of non-co-localized
        objects.
    """
    
    # Store the indicies of non-co-localized objects
    all_non_coloc_obj = []

    # count the number of objects
    # Make sure to convert type to int
    n_obj = int(np.max(obj_map))
    #print(n_obj)

    # Get indicies of non-co-localized objects
    # Object index starts from 1 to n_obj+1.
    # The background is 0.
    for idx_obj in range(1, n_obj+1):

        # Only include objects that are not co-localized
        if idx_obj not in all_coloc_obj:
            all_non_coloc_obj.append(idx_obj)

    all_non_coloc_obj = np.array(all_non_coloc_obj)
    #print(len(all_non_coloc_obj))
    
    return all_non_coloc_obj


def make_cell_state_obj_map(obj_map, all_coloc_obj):
    """
    Create 2 object maps cell_state_obj_map and coloc_obj_map.
    
    Parameters
    ----------
    obj_map : numpy.ndarray
        A 3D image of object map in the shape of (N, H, W), where N is
        the number of z-slices. The object map is produced by 3D
        object counter in ImageJ. In the image, the background is 0, 
        and objects are labeled by pixels with values equal to its
        index. All pixel values are unsigned int, which can be 8-bit,
        16-bit, or others, depending on the the output format set
        by ImageJ
        
    all_coloc_obj : numpy.ndarray
        A 1D array of int that stores the indices of co-localized
        objects.
        
    Returns
    -------
    cell_state_obj_map : numpy.ndarray, dtype="uint8" 
        A 3D image of object map in the same shape as the input 
        obj_map. It contains the same objects as the input obj_map,
        but instead of each object labeled with its index, 
        co-localized objects are labeled as 2, non-co-loc objects
        are labeled as 1, and the background is labeled as 0.
        
    coloc_obj_map : numpy.ndarray, dtype="uint8" 
        A 3D image of object map in the same shape as the input 
        obj_map. It only contains co-localized objects, which
        are labeled as 2. The background is labeled as 0.
    """
    
    # Create an empty image
    # dtype of uint8 is acceptable because the max value in the image
    # is 2
    cell_state_obj_map = np.full(obj_map.shape, 0, dtype="uint8")
    #cell_state_obj_map
    
    # Get the indices of non-co-localized objects
    all_non_coloc_obj = _get_all_non_coloc_obj(obj_map, all_coloc_obj)

    
    # Label all co-localized objects in the cell state object map
    for idx_obj in all_coloc_obj:
        #print(idx_obj)
        obj_mask = obj_map == idx_obj
        cell_state_obj_map[obj_mask] = 2

    # Co-localized objects will be used for other
    # analysis, so it's returned separately.
    coloc_obj_map = np.copy(cell_state_obj_map)

    # Label all non-co-localized objects in the cell state object map
    for idx_obj in all_non_coloc_obj:
        #print(idx_obj)
        obj_mask = obj_map == idx_obj
        cell_state_obj_map[obj_mask] = 1

    
    return cell_state_obj_map, coloc_obj_map


def make_counts_table(fov_id, name_stain_A, name_stain_B, obj_map_stain_A, all_coloc_obj_stain_A):
    """
    Make a table of counts of co-loc and non-co-loc objects. 
    
    Parameters
    ----------
    fov_id : string
        A string of chip ID and FOV ID
        
    name_stain_A : string
        The name of stain A
        
    name_stain_B : string
        The name of stain B
        
    obj_map_stain_A : numpy.ndarray
        A 3D image of object map in the shape of (N, H, W), where N is
        the number of z-slices. The object map is produced by 3D
        object counter in ImageJ. In the image, the background is 0, 
        and objects are labeled by pixels with values equal to its
        index. All pixel values are unsigned int, which can be 8-bit,
        16-bit, or others, depending on the the output format set
        by ImageJ
        
    all_coloc_obj_stain_A : numpy.ndarray
        A 1D array of int that stores the indices of co-localized
        objects from stain A.
        
    Returns
    -------
    counts : dict
        A dict that contains fov_id, name_stain_A, name_stain_B, 
        count_all_obj_stain_A, count_coloc_obj_stain_A, 
        count_non_coloc_obj_stain_A
    """
    
    counts = {}

    counts["fov_id"] = fov_id
    counts["name_stain_A"] = name_stain_A
    counts["name_stain_B"] = name_stain_B
    
    # count the total number of objects from stain A
    # Make sure to convert type to int
    n_obj_stain_A = int(np.max(obj_map_stain_A))
    #print(n_obj_stain_A)
    counts["count_all_obj_stain_A"] = n_obj_stain_A

    # count the number of co-localized objects from stain A
    count_coloc_obj_stain_A = len(all_coloc_obj_stain_A)
    #print(f"number of co-localized objects in stain A: {count_coloc_obj_stain_A}")
    counts["count_coloc_obj_stain_A"] = count_coloc_obj_stain_A

    # count the number of non-co-localized objects from stain A
    count_non_coloc_obj_stain_A = n_obj_stain_A - count_coloc_obj_stain_A
    #print(f"number of non-co-localized objects in stain A: {count_non_coloc_obj_stain_A}")
    counts["count_non_coloc_obj_stain_A"] = count_non_coloc_obj_stain_A

    #print(counts)
    
    return counts


def coloc_channel_pair(obj_map_stain_A, obj_map_stain_B, fov_id, name_stain_A, name_stain_B, threshold_ratio_overlap, output_dir_path):
    """
    The main wrapper function that does co-localization for a given
    pair of object maps from 2 different stains. There are 3 primary
    sections: co-localizaiton operation, creating co-localization
    maps, and writing outputs to files. 
    
    Parameters
    ----------  
    obj_map_stain_A : numpy.ndarray
        A 3D image of object map in the shape of (N, H, W), where N is
        the number of z-slices. The object map is produced by 3D
        object counter in ImageJ. In the image, the background is 0, 
        and objects are labeled by pixels with values equal to its
        index. All pixel values are unsigned int, which can be 8-bit,
        16-bit, or others, depending on the the output format set
        by ImageJ
        
    obj_map_stain_B : numpy.ndarray
        A 3D image of object map in the shape of (N, H, W), where N is
        the number of z-slices. The object map is produced by 3D
        object counter in ImageJ. In the image, the background is 0, 
        and objects are labeled by pixels with values equal to its
        index. All pixel values are unsigned int, which can be 8-bit,
        16-bit, or others, depending on the the output format set
        by ImageJ
        
    fov_id : string
        A string of chip ID and FOV ID
        
    name_stain_A : string
        The name of stain A
        
    name_stain_B : string
        The name of stain B
    
    threshold_ratio_overlap : float
        The minimum ratio of overlap between objects to be considered
        as co-localized
        
    output_dir_path : string
        The path to the output directory
        
    Returns
    -------
    counts : dict
        A dict that contains fov_id, name_stain_A, name_stain_B, 
        count_all_obj_stain_A, count_coloc_obj_stain_A, 
        count_non_coloc_obj_stain_A
    """
    
    
    ## Co-localization operation
    all_coloc_stats, all_coloc_obj_stain_B = fractional_overlap(
        obj_map_stain_A, obj_map_stain_B, threshold_ratio_overlap
    )

    # Turn stats into a data frame
    all_coloc_stats = pd.DataFrame(all_coloc_stats, index=None)
    display(all_coloc_stats)

    # Get the indices of co-localized stain A objects
    if len(all_coloc_stats) > 0:
        all_coloc_obj_stain_A = all_coloc_stats["idx_obj_stain_A"].to_numpy()
    # If there are no co-loc objects, set all_coloc_obj_stain_A to an 
    # empty array. This prevents an error because all_coloc_stats is 
    # empty and there's no key "idx_obj_stain_A"
    else:
        all_coloc_obj_stain_A = np.array([])

    # Make a counts table from the stats
    counts = make_counts_table(
        fov_id, name_stain_A, name_stain_B,
        obj_map_stain_A, all_coloc_obj_stain_A
    )

    ### further analysis
    # Stain A objects that co-localize with more than 1 stain B objects
    #display(all_coloc_stats[all_coloc_stats["n_coloc_obj_stain_B"] == 2])

    # Sort the overlap ratio in increasing order
    #display(np.sort(np.array(all_coloc_stats["ratio_overlap"])))

    # Find the max overlap ratio
    #display(all_coloc_stats[all_coloc_stats["ratio_overlap"] == all_coloc_stats["ratio_overlap"].max()])

    ### Visual inspection
    #obj_mask_stain_A = obj_map_stain_A == 106
    #display_img(obj_mask_stain_A[16])

    #obj_mask_stain_B = obj_map_stain_B == 10
    #display_img(obj_mask_stain_B[16])


    ## Create maps
    ## Stain co-localization map
    stain_coloc_map_rgb = make_stain_coloc_map(
        obj_map_stain_A, obj_map_stain_B,
        all_coloc_obj_stain_A, all_coloc_obj_stain_B
    )
    print(stain_coloc_map_rgb.shape)

    ## Cell state object map
    cell_state_obj_map, coloc_obj_map = make_cell_state_obj_map(
        obj_map_stain_A, all_coloc_obj_stain_A
    )

    # Change label from 1 to 255
    cell_state_obj_map = cell_state_obj_map * 127
    coloc_obj_map = coloc_obj_map * 127

    ##### sanity check
    #print(np.unique(cell_state_obj_map))
    #print(np.unique(coloc_obj_map))

    # Display maps
    #print("stain_coloc_map_rgb:")
    #display_img(stain_coloc_map_rgb[15])
    #print("cell_state_obj_map")
    #display_img(cell_state_obj_map[15])
    #print("coloc_obj_map")
    #display_img(coloc_obj_map[15])
    #####

    
    ## Write outputs to files
    # Save co-localization stats to a csv file
    output_category = ""
    file_name = f"{name_stain_A}_{name_stain_B}_coloc_stats.csv"
    df = all_coloc_stats
    df_index = False
    data_frame_to_csv(output_dir_path, output_category, file_name, df, df_index)

    # Save the stain co-localization map to a TIFF image
    file_name = f"{name_stain_A}_{name_stain_B}_stain_coloc_map.tif"
    img = stain_coloc_map_rgb
    photometric = 'rgb'
    write_tiff(output_dir_path, file_name, img, photometric)

    # Save the cell state object map to a TIFF image
    file_name = f"{name_stain_A}_cell_state_obj_map.tif"
    img = cell_state_obj_map
    write_tiff(output_dir_path, file_name, img)

    # Save the co-localized object map to a TIFF image
    file_name = f"{name_stain_A}_coloc_obj_map.tif"
    img = coloc_obj_map
    write_tiff(output_dir_path, file_name, img)
    
    print("Counts:")
    display(pd.DataFrame([counts]))

    
    return counts

