"""
feat_analysis.py

Description: 
    Functions that analyzes extracted features.
    
"""


__author__ = "John Lin"
__contact__ = "john.lin@emulatebio.com"
__date__ = "2023-05-28"
__status__ = "Prototype"


import pandas as pd

from utils import print_dash_line
from utils import data_frame_to_csv


def display_all_feat(all_feat, output_dir_path):
    """
    Display tables of features of all classes of all FOV of all chips.
    
    Parameters
    ----------
    all_feat : list
        A list of Features object, which stores features of 
        every FOV. Therefore, the dict contains features of all
        FOV and all chips.
        
    output_dir_path : string
        The path to the output directory
    """
    
    for fov_feat in all_feat:
        display(fov_feat.chip_id)
        display(fov_feat.fov_id)
        print(f"percent_migrated: {fov_feat.percent_migrated}")
        print(f"count_migrated: {fov_feat.count_migrated}")
        display(fov_feat.class_feat_table)
        
        # Write table of features to a csv file
        output_category = "all_class_features_all_fov"
        file_name = f"{fov_feat.chip_id}_{fov_feat.fov_id}_class_features.csv"
        df = fov_feat.class_feat_table
        df_index = True
        data_frame_to_csv(
            output_dir_path, output_category, file_name, df, df_index
        )
        
        
def make_feature_summary(all_feat, feature_name, output_dir_path):
    """
    Create a table of summary only for the specified features across
    all chips and FOV.
    
    Parameters
    ----------
    all_feat : list
        A list of Features object, which stores features of 
        every FOV. Therefore, the dict contains features of all
        FOV and all chips.
        
    feature_name : string
        Specify which feature to create a summary for.
        Match one of n_obj, total_volume_obj, mean_intensity_obj,
        or sum_intensity_obj.
        
    output_dir_path : string
        The path to the output directory
        
    Returns
    -------
    feat_summary : pandas.DataFrame
        A table showing features of all classes in columns, and 
        of all FOV and chips in rows.
    """
    
    print(feature_name)
    
    # A list that stores the summary of the features for all 
    # chip ID and FOV ID
    feat_summary = []

    # Loop over features of all FOVs
    for fov_feat in all_feat:

        class_feat_table = fov_feat.class_feat_table

        # Extract the column of the specified feature
        feat_series = class_feat_table[feature_name]
        feat_series = pd.Series(feat_series, copy=True)

        # Add the feature name to the column name of the output table
        feat_series.index = feature_name + " " + feat_series.index

        # Add chip ID, FOV ID, and % migration
        other_info = {
            "chip_id": fov_feat.chip_id,
            "fov_id": fov_feat.fov_id,
            "% migrated": fov_feat.percent_migrated,
            "count migrated": fov_feat.count_migrated
        }
        other_info = pd.Series(other_info)

        # Concatenate the two series
        feat_series = pd.concat([other_info, feat_series])

        #print(feat_series)
        #print("\n")

        feat_summary.append(feat_series)

    # Turn into a data frame
    feat_summary = pd.DataFrame(feat_summary)
    
    # Write to an excel file
    #output_category = "feature_summary"
    #file_name = f"{feature_name}_summary.csv"
    #df = feat_summary
    #df_index = False
    #data_frame_to_csv(
    #    output_dir_path, output_category, file_name, df, df_index
    #)
    
    return feat_summary


def feature_analysis(all_feat, output_dir_path):
    """
    The main function that does feature analysis for all FOV in all
    chips of a given channel.
    
    The outputs consist of all extracted features, feature summaries,
    and all other requested analysis outputs.
    
    Parameters
    ----------
    all_feat : list
        A list of Features object, which stores features of 
        every FOV. Therefore, the dict contains features of all
        FOV and all chips.
        
    output_dir_path : string
        The path to the output directory
        
    Returns
    -------
    n_obj_summary : pandas.DataFrame
        A table showing n_obj of all classes in columns, and 
        of all FOV and chips in rows.
    
    total_volume_obj_summary : pandas.DataFrame
        A table showing total_volume of all classes in columns, and 
        of all FOV and chips in rows.
    
    mean_intensity_obj_summary : pandas.DataFrame
        A table showing mean_intensity of all classes in columns, and 
        of all FOV and chips in rows.
    
    sum_intensity_obj_summary : pandas.DataFrame
        A table showing sum_intensity of all classes in columns, and 
        of all FOV and chips in rows.
    """
    
    ## All extracted features
    print_dash_line()
    print("ALL EXTRACTED FEATURES")

    display_all_feat(all_feat, output_dir_path)

    print_dash_line()


    ## Summaries of features
    print_dash_line()
    print("SUMMARIES OF FEATURES")

    # Create a feature summary table for n_obj
    feature_name = "n_obj"
    n_obj_summary = make_feature_summary(all_feat, feature_name, output_dir_path)
    display(n_obj_summary)

    # Create a feature summary table for total_volume_obj
    feature_name = "total_volume_obj"
    total_volume_obj_summary = make_feature_summary(all_feat, feature_name, output_dir_path)
    display(total_volume_obj_summary)

    # Create a feature summary table for mean_intensity_obj
    feature_name = "mean_intensity_obj"
    mean_intensity_obj_summary = make_feature_summary(all_feat, feature_name, output_dir_path)
    display(mean_intensity_obj_summary)

    # Create a feature summary table for sum_intensity_obj
    feature_name = "sum_intensity_obj"
    sum_intensity_obj_summary = make_feature_summary(all_feat, feature_name, output_dir_path)
    display(sum_intensity_obj_summary)
    
    
    print_dash_line()
    
    ## n_obj total exclude below membrane per FOV
    print("n_obj total exclude below membrane per FOV")
    n_obj_summary_temp = pd.DataFrame(n_obj_summary, copy=True)
    
    # Calculate the number of objects excluding the bottom channel
    n_obj_summary_temp["n_obj total exclude below membrane"] = (
        n_obj_summary_temp["n_obj on top and above membrane"] + 
        n_obj_summary_temp["n_obj within membrane"] + 
        n_obj_summary_temp["n_obj at the bottom of membrane"]
    )
    #display(n_obj_summary_temp)
    
    # Make a copy of the column and rename it
    n_obj_total_exclude_below_membrane_FOV = pd.DataFrame(n_obj_summary_temp[["chip_id", "fov_id", "n_obj total exclude below membrane"]], copy=True)
    display(n_obj_total_exclude_below_membrane_FOV)
    
    print_dash_line()


    ## percentage of intensity of objects in the top channel per FOV
    print_dash_line()
    print("% INTENSITY IN THE TOP CHANNEL PER FOV")

    # the percentage is the sum intensity in the top channel divided by
    # total sum intensity except the bottom channel
    # (at the top and above membrane) / (at the bottom + inside
    # membrane + at the top and above membrane)
    # The intensity below the membrane (aka in the bottom channel)
    # is excluded.
    percent_intensity_top_channel = pd.DataFrame(sum_intensity_obj_summary[["chip_id", "fov_id"]], copy=True)
    percent_intensity_top_channel["% sum_intensity_obj on top and above membrane"] = sum_intensity_obj_summary["sum_intensity_obj on top and above membrane"] / (
        sum_intensity_obj_summary["sum_intensity_obj on top and above membrane"] + 
        sum_intensity_obj_summary["sum_intensity_obj within membrane"] + 
        sum_intensity_obj_summary["sum_intensity_obj at the bottom of membrane"]
    ) * 100

    display(percent_intensity_top_channel)
    
    print_dash_line()
    
    
    # Concatenate all FOV summaries and features and write to a file   
    # The purpose of .iloc[:,4:] is to prevent duplicating the first 4 
    # columns, which are the same in every FOV summary table
    fov_features = pd.concat([
        n_obj_summary, 
        total_volume_obj_summary.iloc[:,4:],
        mean_intensity_obj_summary.iloc[:,4:],
        sum_intensity_obj_summary.iloc[:,4:],
        n_obj_total_exclude_below_membrane_FOV["n_obj total exclude below membrane"],
        percent_intensity_top_channel["% sum_intensity_obj on top and above membrane"]
    ], axis=1, copy=True)
    #display(fov_features)
    
    # Write to an excel file
    output_category = ""
    file_name = f"fov_features.csv"
    df = fov_features
    df_index = False
    data_frame_to_csv(
        output_dir_path, output_category, file_name, df, df_index
    )
    
    
    ## intensity of objects in the top channel per FOV
    # Highlight the mean_intensity_obj and sum_intensity_obj features
    # in the top channel
    intensity_fov_top_channel = pd.DataFrame(mean_intensity_obj_summary[["chip_id", "fov_id", "mean_intensity_obj on top and above membrane"]], copy=True)
    intensity_fov_top_channel["sum_intensity_obj on top and above membrane"] = sum_intensity_obj_summary[["sum_intensity_obj on top and above membrane"]]
    #display(intensity_fov_top_channel)
    
    
    ## intensity of objects in the top channel per chip
    print_dash_line()
    print("INTENSITY IN THE TOP CHANNEL PER CHIP")

    # Take the mean and sum of mean_intensity_obj in the top channel by chip_id
    mean_intensity_chip_top_channel = intensity_fov_top_channel.groupby("chip_id")["mean_intensity_obj on top and above membrane"].agg(["mean", "sum"])

    # Rename the columns
    mean_intensity_chip_top_channel.columns = [
        "mean of mean_intensity_obj on top and above membrane",
        "sum of mean_intensity_obj on top and above membrane"
    ]

    #display(mean_intensity_chip_top_channel)

    # Take the mean and sum of sum_intensity_obj in the top channel by chip_id
    sum_intensity_chip_top_channel = intensity_fov_top_channel.groupby("chip_id")["sum_intensity_obj on top and above membrane"].agg(["mean", "sum"])

    # Rename the columns
    sum_intensity_chip_top_channel.columns = [
        "mean of sum_intensity_obj on top and above membrane",
        "sum of sum_intensity_obj on top and above membrane"
    ]

    #display(sum_intensity_chip_top_channel)

    # Concatenate the data frames
    intensity_chip_top_channel = pd.concat([mean_intensity_chip_top_channel, sum_intensity_chip_top_channel], axis=1)
    display(intensity_chip_top_channel)

    print_dash_line()


    ## total count of migrated objects per Chip
    print_dash_line()
    print("MIGRATED OBJECTS PER CHIP")

    # Sum the migrated count and total count by chip ID
    migrated_per_chip = n_obj_summary_temp.groupby("chip_id")[["count migrated","n_obj total exclude below membrane"]].sum()
    
    # The percentage of migrated objects per Chip
    # Divide the migrated count by the the number of objects in total
    # excluding below membrane to get the percentage.
    # (inside membrane + at the top and above membrane) / (at the bottom
    # + inside membrane + at the top and above membrane)
    # The counts below the membrane (aka bottom channel) are excluded.
    migrated_per_chip["percent count migrated"] = migrated_per_chip["count migrated"] / migrated_per_chip["n_obj total exclude below membrane"] * 100
    display(migrated_per_chip)
        
    chip_features = pd.concat([
        intensity_chip_top_channel,
        migrated_per_chip
    ], copy=True, axis=1)
    #display(chip_features)
    
    # Write to a csv file
    output_category = ""
    file_name = f"chip_features.csv"
    df = chip_features
    df_index = True
    data_frame_to_csv(output_dir_path, output_category, file_name, df, df_index)
    
    print_dash_line()
    
    return n_obj_summary, total_volume_obj_summary, mean_intensity_obj_summary, sum_intensity_obj_summary