"""
feat_extract.py

Description: 
    Functions that identifies key object features.
    
"""


__author__ = "John Lin"
__contact__ = "john.lin@emulatebio.com"
__date__ = "2023-05-28"
__status__ = "Prototype"


import pandas as pd

from unit_tests import check_id_uniqueness, check_index_consistency, check_pos_uniqueness, check_feat_consistency
from classification import classify_obj


class Features:
    """
    A container that stores the features of all classes in a FOV.
    
    Attributes
    ----------
    chip_id : string
        chip_id
        
    fov_id : string
        fov_id
        
    class_feat_table : list
        A list of dictionaries that store features for all classes.
        
    percent_migrated : float
        The percentage is calculated as the counts (above the 
        membrane + within the membrane) / (above the membrane
        + within the membrane + at the bottom of the membrane)
        The denominator excludes counts below the membrane
        (aka counts in the bottom channel)
        
    count_migrated : int
        Counts (above the membrane + within the membrane)
        
    class_feat_index : list
        A list of class names.
        
    Methods
    -------
    add_feat(class_feat)
    
    to_pd_dataframe()
    """
    
    def __init__(self):
        
        self.chip_id = ""
        self.fov_id = ""
        
        # A list of dictionaries that store features for all classes
        self.class_feat_table = []
        self.percent_migrated = 0
        self.count_migrated = 0
        
        self.class_feat_index = [
            "overall FOV",
            "below membrane", 
            "at the bottom of membrane", 
            "within membrane", 
            "on top and above membrane"
        ]
        
        
    def add_feat(self, class_feat):
        """
        Add features of each class to internal table class_feat_table
        
        Parameters
        ----------
        class_feat : dict
            Contains number of objects, total volume of all objects, 
            mean intensity of all objects, and sum intensity of all
            objects in the given data frame. 
            It's returned by _feat_extract for every class
        """
        
        self.class_feat_table.append(class_feat)
        
        
    def to_pd_dataframe(self):
        """
        Once features of all classes have been added to the internal
        table class_feat_table, it's converted to a pandas data frame
        for visualization.
        """
        
        self.class_feat_table = pd.DataFrame(self.class_feat_table)
    
        # Set row names
        self.class_feat_table.index = self.class_feat_index
        
        
def calc_migration(fov_feat):
    """
    Calculate the count and percentage of migrated objects
    
    Parameters
    ----------
    fov_feat : Features
        A container that stores the features of all classes in a FOV.
    """
    
    class_feat_table = fov_feat.class_feat_table
    #class_feat_table

    n_above = class_feat_table["n_obj"]["on top and above membrane"]
    n_within = class_feat_table["n_obj"]["within membrane"]
    n_bottom = class_feat_table["n_obj"]["at the bottom of membrane"]

    percent_migrated = 0

    # Only do the calculation when the denominator is non-zero.
    # Otherwise, a division by zero error would occur
    if n_above + n_within + n_bottom > 0:
        
        # The percentage is calculated as the counts (above the 
        # membrane + within the membrane) / (above the membrane
        # + within the membrane + at the bottom of the membrane)
        # The denominator excludes counts below the membrane,
        # aka counts in the bottom channel)
        percent_migrated = (n_above + n_within) / (n_above + n_within + n_bottom) * 100
        percent_migrated = round(percent_migrated, 2)

    #print(percent_migrated)
    fov_feat.percent_migrated = percent_migrated
    fov_feat.count_migrated = n_above + n_within
    
    
def _feat_extract(df):
    """
    Calculate the features of objects from the given data frame.
    Called for every class.
    
    Parameters
    ----------
    df : numpy.DataFrame
        a subset of the full statistics table produced by the ImageJ
        script. Theoretically, it should contain only 1 unique Chip ID
        and 1 unique FOV ID. 
        Depending on the case, df can be specific to a particular 
        class of objects (above, within, bottom, below)
        
    Returns
    -------
    features : dict
        contains number of objects, total volume of all objects, 
        mean intensity of all objects, and sum intensity of all
        objects in the given data frame
        
    """
    
    
    # Number of objects
    n_obj = len(df)
    #print(n_obj)

    # Total volume of all objects
    total_volume_obj = df["Volume (micron^3)"].sum()
    #print(total_volume_obj)
    
    # Sum intensity of all objects
    # The IntDen field, according to the source code, is the 
    # summed intensity of the object
    # The sum_intensity_obj is the summed intensity of all objects
    sum_intensity_obj = df["IntDen"].sum()

    # Mean intensity of all objects
    if n_obj == 0:
        mean_intensity_obj = 0
    else:
        # The IntDen field, according to the source code, is the 
        # summed intensity of the object
        # The mean intensity is the sum of all pixel intensities
        # of all objects, divided by the sum of all object volume
        mean_intensity_obj = sum_intensity_obj / df["Nb of obj. voxels"].sum()
        
    #print(mean_intensity_obj)
    
    features = {
        "n_obj": n_obj,
        "total_volume_obj": total_volume_obj,
        "mean_intensity_obj": mean_intensity_obj,
        "sum_intensity_obj": sum_intensity_obj
    }

    return features


def feat_extract_all_class(subset_df, class_idx):
    """
    A wrapper function that handles all classes of objects 
    for _feat_extract. Called for every FOV.
    
    Parameters
    ----------
    subset_df : numpy.DataFrame
        a data frame which is a subset of the full statistics table
        produced by the ImageJ script. Theoretically, it should 
        contain only 1 unique Chip ID and 1 unique FOV ID.
        
    class_idx : list of list
        contains indices that specifies which rows in the subset_df
        belong to a particular class of objects (above, within, 
        bottom, below)
    
    Returns
    -------
    fov_feat : Features
        contains the chip_id, fov_id, percent_migrated,
        count_migrated, and a table of features for all classes
        in the FOV, each in its row. 
        
    """
    
    # A Features object that stores features for all classes
    fov_feat = Features()
    
    # Features for overall data frame
    class_feat = _feat_extract(subset_df)
    fov_feat.add_feat(class_feat)
    
    # Extract features for each class of objects
    for i in range(len(class_idx)):
        
        # Use a list of indices to subset the data frame
        class_i_idx = class_idx[i]
        subset_df_class = subset_df.iloc[class_i_idx]            
        #print(subset_df_class)

        # Call _feat_extract function on the subset
        class_feat = _feat_extract(subset_df_class)
        fov_feat.add_feat(class_feat)
    
    # Convert to DataFrame
    fov_feat.to_pd_dataframe()
    
    # Calculate percent_migrated
    calc_migration(fov_feat)
    
    return fov_feat


def feat_extract_all_chips_per_channel(file_path, memb_est_all_chips, output_dir_path):
    """
    The main function that does classification and feature extraction
    for all FOV in all chips of a given channel.
    
    Parameters
    ----------
    file_path : string
        The path to the excel spreadsheet that contains all
        statistics of all objects
        
    memb_est_all_chips : dict
        Store the coefficients of membrane estimates using
        MembraneEstimate objects as containers. The keys are chip_id.
        Returned by estimate_memb_all_chips().
        
    output_dir_path : string
        The path to the output directory
    
    Returns
    -------
    all_feat : list
        A list of Features object, which stores features of 
        every FOV. Therefore, the dict contains features of all
        FOV and all chips.
    """

    # Load the excel file
    xls = pd.ExcelFile(file_path)

    # Data of different chip ID are stored in different sheets
    sheet_names = xls.sheet_names

    # A list that stores features of all chip ID and FOV ID
    all_feat = []

    # Perform analysis for different chip ID
    for sheet in range(0, len(xls.sheet_names)):

        chip_id = xls.sheet_names[sheet]
        print(chip_id)

        df_chip = pd.read_excel(xls, chip_id, header=1)
        
        # Use the previously calculated membrane estimates
        # Estimate the membrane shape and location using all FOV
        # in the same chip. In classification, all FOV in the
        # same chip use the same fit coefficients
        mem_est = memb_est_all_chips[chip_id]
        p = mem_est.p
        pp = mem_est.pp
        
        # Get a list of unique FOV ID
        unique_fov_id = df_chip["FOV ID"].unique()
        print(unique_fov_id)

        #print(df_chip[["Chip ID", "FOV ID"]])

        # Perform analysis for different FOV ID
        for fov_id in unique_fov_id:

            print(f"fov_id: {fov_id}")

            # Subset the data frame by FOV ID
            subset_df_fov = df_chip[df_chip["FOV ID"] == fov_id]
            #print(len(subset_df_fov))
            #print(subset_df_fov)

            # Unit tests
            check_id_uniqueness(subset_df_fov)

            display(p)
            display(pp)
        
            # Classification
            class_pos, class_idx = classify_obj(
                subset_df_fov, chip_id, fov_id, p, pp, output_dir_path
            )

            # Unit tests
            check_index_consistency(subset_df_fov, class_idx, class_pos)
            check_pos_uniqueness(subset_df_fov, class_pos)


            # Extract features (separately for all classes of objects)
            fov_feat = feat_extract_all_class(subset_df_fov, class_idx)
            #print(fov_feat)
            
            # Store the features of each FOV
            fov_feat.chip_id = chip_id
            fov_feat.fov_id = "fov_" + str(fov_id)
            all_feat.append(fov_feat)
            
            # Unit tests
            check_feat_consistency(fov_feat.class_feat_table)
            
    
    return all_feat