"""
gui.py

Description: 
    GUI functions
    
"""


__author__ = "John Lin"
__contact__ = "john.lin@emulatebio.com"
__date__ = "2023-05-28"
__status__ = "Prototype"


import PySimpleGUI as sg
import os


def gui_feat_analysis():
    """
    Open a dialog window for user to enter parameters.
    
    Returns
    -------
    run_analysis : bool
        Indicate whether the user wants to run the analysis
        
    min_Z_memb : int
        The lowest point on the z-axis, in micron, where the membrane
        can be possibly found.
        
    max_Z_memb : int
        The highest point on the z-axis, in micron, where the membrane
        can be possibly found.
        
    file_path_cell_tracker : string
        The path to the excel spreadsheet that contains all
        statistics of all objects of cell tracker CAR-T channel
        
    file_path_caspase : string
        The path to the excel spreadsheet that contains all
        statistics of all objects of the caspase channel
        
    output_dir_root : string
        File path to the root of the output directory
    
    project_name : string
        The name of the project
    """
    
    # Initialize variables
    # Range of coordinate in z-axis of lower boundary of membrane
    min_Z_memb = 0
    max_Z_memb = 0
    
    # File paths to excel spreadsheets
    file_path_cell_tracker = ""
    file_path_caspase = ""
    
    # File path to the root of the output directory
    output_dir_root = ""
    
    project_name = ""
    
    # Number of fields in the GUI
    field_counts = 6
    
    # Indicate whether the user wants to start the analysis
    run_analysis = False
    
    
    # Create GUI layout
    sg.theme('SystemDefault1')
    
    layout = [
        [sg.Text('Expected range of coordinate in z-axis of lower boundary of membrane.')],
        [sg.Text('Minimum (uM)', size=(18, 1)), sg.InputText("100")],
        [sg.Text('Maximum (uM)', size=(18, 1)), sg.InputText("300")],
        [sg.Text('')],
        [sg.Text('Open the Excel spreadsheet of object statistics of the following channel of stains')],
        [sg.Text('Cell tracker for CAR-T', size=(18, 1)), sg.Input(), sg.FileBrowse()],
        [sg.Text('Caspase', size=(18, 1)), sg.Input(), sg.FileBrowse()],
        [sg.Text('')],
        [sg.Text('Select output directory')],
        [sg.Text('Output folder', size=(18, 1)), sg.InputText("../output"), sg.FolderBrowse()],
        [sg.Text('')],
        [sg.Text('Other info')],
        [sg.Text('Project name', size=(18, 1)), sg.InputText("Project")],
        [sg.OK(), sg.Cancel()],
    ]

    # Run GUI
    window = sg.Window('Input parameters', layout, enable_close_attempted_event=True, finalize=True)
    window.bring_to_front()
    
    while True:

        event, values = window.read()

        if event == "OK":

            # Check if all input fields are filled out
            input_complete = True
            
            # This is used instead of for "key in values.keys():"
            # because values.keys() contains "browse", "browse0", etc.
            # which may be empty if the user uses the default value
            # or fills out manually.
            for key in range(field_counts):
                #print(key)
                #print(values[key])
                if values[key] == "":
                    input_complete = False

            # If all fields are filled out, read the values and exit the GUI
            if input_complete:
                
                # Range of coordinate in z-axis of lower boundary of membrane
                min_Z_memb = int(values[0])
                max_Z_memb = int(values[1])
                file_path_cell_tracker = values[2]
                file_path_caspase = values[3]
                output_dir_root = values[4]
                project_name = values[5]
                
                # If the directory exists and is not empty, display a
                # warning message
                output_dir_path = f"{output_dir_root}/feature_analysis_output"
                if os.path.isdir(output_dir_path) and os.listdir(output_dir_path):
                    
                    # If the user wants to proceed, start the analysis
                    # run_analysis becomes True
                    if sg.popup_yes_no('Output directory is not empty. Existing files from previous runs of feature analysis may be overwritten. Proceed?') == 'Yes':
                        run_analysis = True
                        break
                        
                    # If the user does not want to proceed, return
                    # to GUI
                    # run_analysis remains False
                    else:
                        pass
                        
                # If the directory does not exist or exists but empty,
                # start the analysis. No need for warning
                # run_analysis becomes True
                else:
                    run_analysis = True
                    break

            # If some fields are empty, display a message and return to the GUI
            else:
                sg.popup('Please fill out all fields.')

        # Double check with the user if he wants to exit
        # run_analysis remains False
        elif event == sg.WINDOW_CLOSE_ATTEMPTED_EVENT or event == 'Cancel':
            if sg.popup_yes_no('Do you really want to exit?') == 'Yes':
                break

    # Close the GUI
    window.close()

    return run_analysis, min_Z_memb, max_Z_memb, file_path_cell_tracker, file_path_caspase, output_dir_root, project_name


def gui_coloc():
    """
    Open a dialog window for user to enter parameters.
    
    Returns
    -------
    run_analysis : bool
        Indicate whether the user wants to run the analysis
        
    name_stain_A : string
        The name of stain A
        
    name_stain_B : string
        The name of stain B
        
    file_path_stain_A : string
        The filename of the object map of stain A
        
    file_path_stain_B : string
        The filename of the object map of stain B
        
    threshold_ratio_overlap : float
        The minimum ratio of overlap between objects to be considered
        as co-localized
        
    output_dir_root : string
        File path to the root of the output directory
    
    project_name : string
        The name of the project
    """
    
    # Initialize variables
    # Name of the stains
    name_stain_A = ""
    name_stain_B = ""
    
    # File paths to object maps
    file_path_stain_A = ""
    file_path_stain_B = ""
        
    # FOV ID
    #fov_id = ""
    
    # Threshold of overlap ratio
    threshold_ratio_overlap = 0

    # File path to the root of the output directory
    output_dir_root = ""
    
    project_name = ""
    
    # Number of fields in the GUI
    field_counts = 7
    
    # Indicate whether the user wants to start the analysis
    run_analysis = False
    
    
    # Create GUI layout
    sg.theme('SystemDefault1')
    
    layout = [
        [sg.Text('Enter name of stain A. Example: cell_tracker_CAR-T')],
        [sg.Text('Name of stain', size=(18, 1)), sg.InputText("cell_tracker_CAR-T")],
        [sg.Text('Select object map of stain A. Example: Objects map of r01c03f09ch1_compiled.tif')],
        [sg.Text('Path to object map', size=(18, 1)), sg.Input(), sg.FileBrowse()],
        [sg.Text('')],
        [sg.Text('Enter name of stain B. Example: caspase')],
        [sg.Text('Name of stain', size=(18, 1)), sg.InputText("caspase")],
        [sg.Text('Select object map of stain B. Example: Objects map of r01c03f09ch4_compiled.tif')],
        [sg.Text('Path to object map', size=(18, 1)), sg.Input(), sg.FileBrowse()],
        [sg.Text('')],
        #[sg.Text('Enter FOV ID. Example: r01c03f09')],
        #[sg.Text('FOV ID', size=(18, 1)), sg.InputText("r01c03f09")],
        #[sg.Text('')],
        [sg.Text('The minimum ratio of overlap between objects to be considered as co-localized. (Valid range: 0.0 to 1.0)')],
        [sg.Text('Threshold of overlap ratio', size=(18, 1)), sg.InputText()],
        [sg.Text('')],
        [sg.Text('Select output directory')],
        [sg.Text('Output folder', size=(18, 1)), sg.InputText("../output"), sg.FolderBrowse()],
        [sg.Text('')],
        [sg.Text('Other info')],
        [sg.Text('Project name', size=(18, 1)), sg.InputText("Project")],
        [sg.OK(), sg.Cancel()],
    ]

    # Run GUI
    window = sg.Window('Input parameters', layout, enable_close_attempted_event=True, finalize=True)
    window.bring_to_front()
    
    while True:

        event, values = window.read()

        if event == "OK":

            # Check if all input fields are filled out
            input_complete = True
            
            # This is used instead of for "key in values.keys():"
            # because values.keys() contains "browse", "browse0", etc.
            # which may be empty if the user uses the default value
            # or fills out manually.
            for key in range(field_counts):
                #print(key)
                #print(values[key])
                if values[key] == "":
                    input_complete = False

            # If all fields are filled out, read the values and exit the GUI
            if input_complete:
                
                name_stain_A = values[0]
                file_path_stain_A = values[1]
                name_stain_B = values[2]
                file_path_stain_B = values[3]
                #fov_id = values[4]
                threshold_ratio_overlap = float(values[4])
                output_dir_root = values[5]
                project_name = values[6]
                
                run_analysis = True
                break
                
            # If some fields are empty, display a message and return to the GUI
            else:
                sg.popup('Please fill out all fields.')

        # Double check with the user if he wants to exit
        # run_analysis remains False
        elif event == sg.WINDOW_CLOSE_ATTEMPTED_EVENT or event == 'Cancel':
            if sg.popup_yes_no('Do you really want to exit?') == 'Yes':
                break

    # Close the GUI
    window.close()

    return run_analysis, name_stain_A, name_stain_B, file_path_stain_A, file_path_stain_B, threshold_ratio_overlap, output_dir_root, project_name


def gui_max_proj():
    """
    Open a dialog window for user to enter parameters.
    
    Returns
    -------
    run_analysis : bool
        Indicate whether the user wants to run the analysis
        
    input_dir_path : string
        File path to the input directory
        
    output_dir_root : string
        File path to the root of the output directory
    
    project_name : string
        The name of the project
    """
    
    # Initialize variables
    input_dir_path = ""
    output_dir_root = ""
    project_name = ""
    
    # Number of fields in the GUI
    field_counts = 3
    
    # Indicate whether the user wants to start the analysis
    run_analysis = False
    
    
    # Create GUI layout
    sg.theme('SystemDefault1')
    
    layout = [
        [sg.Text('Select a folder of 3D TIFF image stacks as the input directory')],
        [sg.Text('Input directory', size=(18, 1)), sg.InputText(), sg.FolderBrowse()],
        [sg.Text('Output directory', size=(18, 1)), sg.InputText("../output"), sg.FolderBrowse()],
        [sg.Text('')],
        [sg.Text('Other info')],
        [sg.Text('Project name', size=(18, 1)), sg.InputText("Project")],
        [sg.OK(), sg.Cancel()],
    ]

    # Run GUI
    window = sg.Window('Input parameters', layout, enable_close_attempted_event=True, finalize=True)
    window.bring_to_front()
    
    while True:

        event, values = window.read()

        if event == "OK":

            # Check if all input fields are filled out
            input_complete = True
            
            # This is used instead of for "key in values.keys():"
            # because values.keys() contains "browse", "browse0", etc.
            # which may be empty if the user uses the default value
            # or fills out manually.
            for key in range(field_counts):
                #print(key)
                #print(values[key])
                if values[key] == "":
                    input_complete = False

            # If all fields are filled out, read the values and exit the GUI
            if input_complete:
                
                input_dir_path = values[0]
                output_dir_root = values[1]
                project_name = values[2]
                
                # If the output directory exists and is not empty, display a
                # warning message
                output_dir_path = f"{output_dir_root}/max_projection_output"
                if os.path.isdir(output_dir_path) and os.listdir(output_dir_path):
                    
                    # If the user wants to proceed, start the analysis
                    # run_analysis becomes True
                    if sg.popup_yes_no('Output directory is not empty. Existing files from previous runs of maximum projection may be overwritten. Proceed?') == 'Yes':
                        run_analysis = True
                        break
                        
                    # If the user does not want to proceed, return
                    # to GUI
                    # run_analysis remains False
                    else:
                        pass
                        
                # If the directory does not exist or exists but empty,
                # start the analysis. No need for warning
                # run_analysis becomes True
                else:
                    run_analysis = True
                    break

            # If some fields are empty, display a message and return to the GUI
            else:
                sg.popup('Please fill out all fields.')

        # Double check with the user if he wants to exit
        # run_analysis remains False
        elif event == sg.WINDOW_CLOSE_ATTEMPTED_EVENT or event == 'Cancel':
            if sg.popup_yes_no('Do you really want to exit?') == 'Yes':
                break

    # Close the GUI
    window.close()

    return run_analysis, input_dir_path, output_dir_root, project_name


def popup_message(message):
    """
    Open a popup window with the specified message.
    
    Parameters
    ----------
    message : string
        Message to display in the popup window.
    """
    
    sg.theme('SystemDefault1')
    sg.popup(message, keep_on_top=True)
    
    
    