"""
input_checks.py

Description:
    Functions that check the format and validity of the input data.
    
"""


__author__ = "John Lin"
__contact__ = "john.lin@emulatebio.com"
__date__ = "2023-06-10"
__status__ = "Prototype"


import pandas as pd
import numpy as np

from gui import popup_message


def input_checks(min_Z_memb, max_Z_memb, file_path_cell_tracker, file_path_caspase):
    """
    Wrapper function that calls check_df_chip_len, 
    check_Z_memb_range, and check_chip_id_equal for specific tasks
    
    Parameters
    ----------
    min_Z_memb : int
        The lowest point on the z-axis, in micron, where the membrane
        can be possibly found.
        
    max_Z_memb : int
        The highest point on the z-axis, in micron, where the membrane
        can be possibly found.
    
    file_path_cell_tracker : string
        The path to the excel spreadsheet that contains all
        statistics of all objects for cell tracker CAR-T
        
    file_path_caspase : string
        The path to the excel spreadsheet that contains all
        statistics of all objects for caspase
        
    Returns
    -------
    bool
        True if all checks are passeed. False if otherwise.
    """
    
    return check_df_chip_len(file_path_cell_tracker) and check_df_chip_len(file_path_caspase) and check_Z_memb_range(min_Z_memb, max_Z_memb, file_path_cell_tracker) and check_chip_id_equal(file_path_cell_tracker, file_path_caspase)


def check_df_chip_len(file_path):
    """
    Check if the excel spreadsheet contains any sheet (chip ID) with 0
    entries. Stop the analysis if so.
    
    Parameters
    ----------
    file_path : string
        The path to the excel spreadsheet that contains all
        statistics of all objects
        
    Returns
    -------
    bool
        True if check is passed. False if otherwise.
    """
    
    # Load the excel file
    xls = pd.ExcelFile(file_path)

    # Data of different chip ID are stored in different sheets
    for chip_id in xls.sheet_names:
        #print(chip_id)

        df_chip = pd.read_excel(xls, chip_id, header=1)
        #print(len(df_chip))
        
        # Stop all analysis if a chip ID contains 0 entries
        if len(df_chip) == 0:
            
            # Display error message
            message = f"{chip_id} contains 0 entries in {file_path}. \n\nPlease remove {chip_id} sheet from all excel spreadsheets. Analysis aborted."
            print(message)
            popup_message(message)
            
            return False
        
    return True


def check_Z_memb_range(min_Z_memb, max_Z_memb, file_path):
    """
    Check if min_Z_memb and max_Z_memb are a valid range for the 
    z-axis coordinates of objects in the excel spreadsheet.
    Stop the analysis if the range is not valid.
    
    Parameters
    ----------
    min_Z_memb : int
        The lowest point on the z-axis, in micron, where the membrane
        can be possibly found.
        
    max_Z_memb : int
        The highest point on the z-axis, in micron, where the membrane
        can be possibly found.
        
    file_path : string
        The path to the excel spreadsheet that contains all
        statistics of all objects
        
    Returns
    -------
    bool
        True if check is passed. False if otherwise.
    """
    
    # Load the excel file
    xls = pd.ExcelFile(file_path)

    # Data of different chip ID are stored in different sheets
    for chip_id in xls.sheet_names:
        #print(chip_id)

        df_chip = pd.read_excel(xls, chip_id, header=1)
        #print(len(df_chip))
        
        # Stop all analysis if no objects fall between min_Z_memb and max_Z_memb
        Z_IC = np.array(df_chip["Z-pos (micron)"])
        condition = (min_Z_memb < Z_IC) & (Z_IC < max_Z_memb)
        #print(condition)
        #print(np.count_nonzero(condition))
        
        if np.count_nonzero(condition) == 0:
            
            # Display error message
            message = f"Membrane cannot be estimated. There are no objects in {chip_id} between the z-axis range of {min_Z_memb} and {max_Z_memb} in file {file_path}. \n\nPlease adjust min_Z_memb and/or max_Z_memb to reflect the range of Z-pos (micron) field in the {chip_id} sheet. Analysis aborted."
            print(message)
            popup_message(message)
            
            return False
        
    return True


def check_chip_id_equal(file_path_cell_tracker, file_path_caspase):
    """
    Check if chip IDs are equal across all excel spreadsheets. 
    Stop the analysis if not.
    
    Parameters
    ----------    
    file_path_cell_tracker : string
        The path to the excel spreadsheet that contains all
        statistics of all objects for cell tracker CAR-T
        
    file_path_caspase : string
        The path to the excel spreadsheet that contains all
        statistics of all objects for caspase
        
    Returns
    -------
    bool
        True if check is passed. False if otherwise.
    """
    
    # Load the excel file
    xls_cell_tracker = pd.ExcelFile(file_path_cell_tracker)
    xls_caspase = pd.ExcelFile(file_path_caspase)
    
    # check if chip ID are equal across all excel spreadsheets
    if np.array_equal(xls_cell_tracker.sheet_names, xls_caspase.sheet_names):
        return True
    
    else:
        # Display error message
        message = f"Chip ID inconsistent. Please make sure all chip ID are present in all excel spreadsheets and remove any inconsistent ones. Analysis aborted."

        print(message)
        popup_message(message)

        return False
    
    
def match_img_shape(obj_map_stain_A, obj_map_stain_B):
    """
    Check if the dimensions between the two input images are
    equal. Stop the analysis if not so.
    
    Parameters
    ----------
    obj_map_stain_A : numpy.ndarray
        A 3D image of object map in the shape of (N, H, W), where N is
        the number of z-slices. The object map is produced by 3D
        object counter in ImageJ. In the image, the background is 0, 
        and objects are labeled by pixels with values equal to its
        index. All pixel values are unsigned int, which can be 8-bit,
        16-bit, or others, depending on the the output format set
        by ImageJ
        
    obj_map_stain_B : numpy.ndarray
        A 3D image of object map in the shape of (N, H, W), where N is
        the number of z-slices. The object map is produced by 3D
        object counter in ImageJ. In the image, the background is 0, 
        and objects are labeled by pixels with values equal to its
        index. All pixel values are unsigned int, which can be 8-bit,
        16-bit, or others, depending on the the output format set
        by ImageJ
    
    Returns
    -------
    bool
        True if check is passed. False if otherwise.
    """
    
    
    if obj_map_stain_A.shape != obj_map_stain_B.shape:
            
        # Display error message
        message = f"The dimensions of the two object maps don't match. Analysis aborted."
        print(message)
        popup_message(message)
            
        return False
        
    return True


def check_img_dim(file_path, img, dim=3):
    """
    Check if the input image has the specified dimension, e.g., 2D, 3D
    or 4D.
    Skip the image for analysis if not so.
    
    Parameters
    ----------
    file_name : string
        File path to the image
        
    img_stack : numpy.ndarray
        An image with unspecified dimension
        
    dim : int
        The required dimension
    """
    
    if len(img.shape) != dim:
        message = f"The image {file_path} is a {len(img.shape)}D image. Please use a {dim}D image instead. Image skipped. Analysis will continue."
        print(message)
        popup_message(message)
            
        return False
    
    return True

