"""
max_projection.py

Description: 
    Functions that apply maximum intensity projection to a 3D image
    
"""


__author__ = "John Lin"
__contact__ = "john.lin@emulatebio.com"
__date__ = "2023-06-23"
__status__ = "Prototype"


import os

import numpy as np
import pandas as pd

from input_checks import check_img_dim
from utils import display_img, read_tiff, write_tiff, data_frame_to_csv, get_tiff_list, find_fov_id_tiff_stack


def max_proj(img_stack, fov_id, ch_index, output_dir_path):
    """
    Apply maximum intensity projection to the image and sum the 
    intensity. 
    
    Parameters
    ----------
    img_stack : numpy.ndarray
        A 3D image in the shape of (N, H, W).
        
    fov_id : string
        A string that represents the FOV ID. It can be the
        actual FOV ID or the file name if FOV ID is not present.
        
    ch_index : string
        A string that represents the channel index. It can be the
        actual channel index or an empty string if channel index
        is not present.
    
    output_dir_path : string
        The path to the output directory
    """

    # Maximum Intensity Projection
    img_max_proj = np.max(img_stack, axis=0)
    #display_img(img_max_proj)

    # Use 64 bit int to prevent integer overflow
    img_max_proj_int64 = np.array(img_max_proj, dtype=np.int64)
    
    # Sum intensity
    sum_intensity = np.sum(img_max_proj_int64)
    #print(type(sum_intensity))
    print(f"sum_intensity = {sum_intensity}")
    print("\n")

    # Write image to a tiff file
    output_file_name = f"max_projected_{fov_id}{ch_index}.tiff"
    write_tiff(output_dir_path, output_file_name, img_max_proj)

    # Create a row of info for pandas data frame
    sum_intensity_info = make_sum_intensity_info(fov_id, ch_index, sum_intensity)
    
    return sum_intensity_info


def make_sum_intensity_info(fov_id, ch_index, sum_intensity):
    """
    Make a table of sum intensity from max projection
    
    Parameters
    ----------
    fov_id : string
        A string of chip ID and FOV ID
        
    ch_index : string
        The channel index
        
    sum_intensity : int
        Sum intensity from max projection
        
    Returns
    -------
    sum_intensity_info : dict
        A dict that contains fov_id, ch_index, and sum_intensity
    """
    
    sum_intensity_info = {}
    sum_intensity_info["fov_id"] = fov_id
    sum_intensity_info["ch_index"] = ch_index
    sum_intensity_info["sum_intensity"] = sum_intensity
    
    return sum_intensity_info


def max_proj_multi_img(input_dir_path, output_dir_path):
    """
    A wrapper function that applies maximum projection and sum
    intensity to all TIFF 3D image stacks in the input directory.
    
    Parameters
    ----------
    input_dir_path : string
        File path to the input directory
        
    output_dir_root : string
        File path to the root of the output directory
    """   
    
    # A list of dict that contains fov_id, ch_index, and 
    # sum_intensity
    sum_intensity_table = []

    # Get the list of tiff files in the input directory
    tiff_file_list = get_tiff_list(input_dir_path)

    # Repeat maximum projection for all tiff files
    for tiff_file_name in tiff_file_list:

        # Find FOV ID and channel index
        fov_id, ch_index = find_fov_id_tiff_stack(tiff_file_name)
        #print(fov_id, ch_index)

        # Include input directory in the file path
        file_path = os.path.join(input_dir_path, tiff_file_name)

        # Read image
        img_stack = read_tiff(file_path)
        
        # Check format and validity of input data
        # Proceed only if the file exists
        ##if img_stack is not None:
        # Check if the input image is 3D. Skip the image if not.
        pass_input_checks = check_img_dim(file_path, img_stack, dim=3)
        if not pass_input_checks:
            print("\n")
            continue
        
        # Maximum Intensity Projection
        sum_intensity_info = max_proj(img_stack, fov_id, ch_index, output_dir_path)
        sum_intensity_table.append(sum_intensity_info)

    # Turn into a pandas data frame
    sum_intensity_table = pd.DataFrame(sum_intensity_table)
    display(sum_intensity_table)

    # Write table to a csv file
    output_category = ""
    file_name = f"sum_intensity_max_projection.csv"
    df = sum_intensity_table
    df_index = False
    data_frame_to_csv(output_dir_path, output_category, file_name, df, df_index)

