"""
plots.py

Description: 
    Functions that render plots for visualization.
    
"""


__author__ = "John Lin"
__contact__ = "john.lin@emulatebio.com"
__date__ = "2023-05-28"
__status__ = "Prototype"


import matplotlib.pyplot as plt
import numpy as np

from utils import plot_to_png


def plot_feat_between_class(all_feat, output_dir_path):
    """
    Create a bar plot for comparing features across classes,
    for all FOV of all chips.
    
    Parameters
    ----------
    all_feat : list
        A list of Features object, which stores features of 
        every FOV. Therefore, the dict contains features of all
        FOV and all chips.
        
    output_dir_path : string
        The path to the output directory
    """
    
    # Loop over features of all FOV of all chips
    for fov_feat in all_feat:

        chip_id = fov_feat.chip_id
        fov_id = fov_feat.fov_id

        print(chip_id)
        print(fov_id)
        #print(f"percent_migrated: {fov_feat.percent_migrated}")
        class_feat_table = fov_feat.class_feat_table

        print(class_feat_table)

        # Get the feature names
        feature_names = class_feat_table.columns
        #print(feature_names)

        # Loop over all features
        for feature_name in feature_names:
            
            # Create the bar plot
            title = f"{feature_name} of each class in {fov_id} of {chip_id}"
            plt.title(title)
            ax = plt.barh(
                class_feat_table[feature_name].index,
                class_feat_table[feature_name]
            )
            plt.bar_label(ax, label_type='edge')
            #plt.xticks(rotation=45, horizontalalignment="right")
            
            # Write bar plot to an image
            output_category = "bar_plots/features_between_classes"
            file_name = f"{feature_name}_of_classes_in_{fov_id}_{chip_id}.png"
            bbox_inches = "tight"
            plot_to_png(
                output_dir_path, output_category, file_name, plt, bbox_inches
            )
            plt.show()

        print("\n")
        
        
def plot_feat_between_fov(feat_summary, feature_name, output_dir_path):
    """
    Create a bar plot for comparing the specified feature across FOV,
    for all chips.
    
    Parameters
    ----------
    feat_summary : pandas.DataFrame
        A table showing features of all classes in columns, and 
        of all FOV and chips in rows.
        
    feature_name : string
        Specify which feature to plot.
        Match one of n_obj, total_volume_obj, mean_intensity_obj,
        or sum_intensity_obj.
        
    output_dir_path : string
        The path to the output directory
    """
        
    # Loop over all chips
    for chip_id in feat_summary['chip_id'].unique():
        print(chip_id)

        # Take the rows that belong to the chip_id
        feat_chip = feat_summary[feat_summary['chip_id'] == chip_id]
        list_fov_id = feat_chip['fov_id'].to_numpy()

        display(list_fov_id)

        # Take the overall FOV column
        summary_feat_fov = feat_chip[f"{feature_name} overall FOV"].to_numpy()

        print(summary_feat_fov)

        # Create the bar plot
        title = f"{feature_name} of each fov in {chip_id}"
        plt.title(title)
        ax = plt.bar(list_fov_id, summary_feat_fov)
        plt.bar_label(ax, label_type='edge')
        
        # Write bar plot to an image
        output_category = "bar_plots/features_between_fov"
        file_name = f"{feature_name}_of_fov_in_{chip_id}.png"
        plot_to_png(output_dir_path, output_category, file_name, plt)
        plt.show()
        
        
def plot_feat_between_chip(feat_summary, feature_name, output_dir_path):
    """
    Create a bar plot for comparing the specified feature across chips.
    
    Parameters
    ----------
    feat_summary : pandas.DataFrame
        A table showing features of all classes in columns, and 
        of all FOV and chips in rows.
        
    feature_name : string
        Specify which feature to plot.
        Match one of n_obj, total_volume_obj, mean_intensity_obj,
        or sum_intensity_obj.
        
    output_dir_path : string
        The path to the output directory
    """
    
    # Get the FOV ID
    list_fov_id = feat_summary['chip_id'].unique()
    print(list_fov_id)

    mean_feat = []
    sum_feat = []

    # Loop over all chips
    for chip_id in feat_summary['chip_id'].unique():
        #print(chip_id)

        # Take the rows that belong to the chip_id
        feat_chip = feat_summary[feat_summary['chip_id'] == chip_id]

        # Take the overall FOV column
        summary_feat_fov = feat_chip[f"{feature_name} overall FOV"].to_numpy()

        # Take the mean and sum across chips
        mean_feat.append(np.mean(summary_feat_fov))
        sum_feat.append(np.sum(summary_feat_fov))

    print(mean_feat)
    print(sum_feat)

    # Create the bar plot using the mean of features
    title = f"mean of {feature_name} across fov of each chip"
    plt.title(title)
    #plt.bar(all_chip_stats.keys(), feature_chip, yerr = 10, ecolor='black', capsize=10)
    ax = plt.bar(list_fov_id, mean_feat)
    plt.bar_label(ax, label_type='edge')
    
    # Write bar plot to an image
    output_category = "bar_plots/features_between_chip"
    file_name = f"mean_of_{feature_name}_of_chips.png"
    plot_to_png(output_dir_path, output_category, file_name, plt)
    plt.show()
    
    # Create the bar plot using the sum of features
    title = f"sum of {feature_name} across fov of each chip"
    plt.title(title)
    #plt.bar(all_chip_stats.keys(), feature_chip, yerr = 10, ecolor='black', capsize=10)
    ax = plt.bar(list_fov_id, sum_feat)
    plt.bar_label(ax, label_type='edge')
    
    # Write bar plot to an image
    output_category = "bar_plots/features_between_chip"
    file_name = f"sum_of_{feature_name}_of_chips.png"
    plot_to_png(output_dir_path, output_category, file_name, plt)
    plt.show()
    
    
def stacked_bar_plot_between_fov(feat_summary, feature_name, output_dir_path):
    """
    Create a stacked bar plot for comparing the specified feature
    across classes and across FOV, for all chips. All classes are
    stacked in the same FOV.
    
    Parameters
    ----------
    feat_summary : pandas.DataFrame
        A table showing features of all classes in columns, and 
        of all FOV and chips in rows.
        
    feature_name : string
        Specify which feature to plot.
        Match one of n_obj, total_volume_obj, mean_intensity_obj,
        or sum_intensity_obj.
        
    output_dir_path : string
        The path to the output directory
    """
        
    # Get the chip ID
    unique_chip_id = feat_summary['chip_id'].unique()
    
    # Loop over all chips
    for chip_id in unique_chip_id:

        # Take the rows that belong to the chip_id
        df_plot = feat_summary[feat_summary['chip_id'] == chip_id]
        
        # Take the columns of fov_id and classes
        df_plot = df_plot[[
            "fov_id",
            f"{feature_name} below membrane",
            f"{feature_name} at the bottom of membrane",
            f"{feature_name} within membrane",
            f"{feature_name} on top and above membrane"
        ]]
        #display(df_plot)

        # Create the stacked bar plot
        ax = df_plot.plot(
            x='fov_id', kind='bar', stacked=True,
            title=f'{feature_name} in {chip_id}'
        )
        for i in ax.containers:
            ax.bar_label(i, label_type='center')
            
        # Write stacked bar plot to an image
        output_category = "bar_plots/stacked_features_between_fov"
        file_name = f"{feature_name}_of_fov_in_{chip_id}.png"
        plot_to_png(output_dir_path, output_category, file_name, plt)
        plt.show()
        
        
def stacked_bar_plot_between_chips(feat_summary, feature_name, output_dir_path):
    """
    Create a stacked bar plot for comparing the specified feature
    sum across FOV for all classes, between chips. All classes
    are stacked in the same chip. The values of the classes are 
    summed across all FOV in the chip.
    
    Parameters
    ----------
    feat_summary : pandas.DataFrame
        A table showing features of all classes in columns, and 
        of all FOV and chips in rows.
        
    feature_name : string
        Specify which feature to plot.
        Match one of n_obj, total_volume_obj, mean_intensity_obj,
        or sum_intensity_obj.
        
    output_dir_path : string
        The path to the output directory
    """

    print("Tables displayed for validation only")
    display(feat_summary)
    
    # Sum the values of all classes across FOV
    df_plot = feat_summary.groupby("chip_id")[[
        f"{feature_name} below membrane",
        f"{feature_name} at the bottom of membrane",
        f"{feature_name} within membrane", 
        f"{feature_name} on top and above membrane"]].sum()
    
    display(df_plot)
    
    # Create a column for the chip_id
    df_plot["chip_id"] = df_plot.index    
    
    # Create the stacked bar plot
    ax = df_plot.plot(
        x="chip_id", kind='bar', stacked=True,
        title=f'sum of {feature_name} across all FOV in a chip'
    )
    for i in ax.containers:
        ax.bar_label(i, label_type='center')
        
    # Write stacked bar plot to an image
    output_category = "bar_plots/stacked_features_between_chips"
    file_name = f"sum_{feature_name}_across_all_FOV_in_chip.png"
    plot_to_png(output_dir_path, output_category, file_name, plt)
    plt.show()
    
    
def make_bar_plots(all_feat, n_obj_summary, total_volume_obj_summary, mean_intensity_obj_summary, sum_intensity_obj_summary, output_dir_path):
    """
    The main plotting function that creates all sets of plots:
    bar plots comparing features between classes, between FOV,
    between chips, and stacked bar plots.
    
    Parameters
    ----------
    all_feat : list
        A list of Features object, which stores features of 
        every FOV. Therefore, the dict contains features of all
        FOV and all chips.
        
    n_obj_summary : pandas.DataFrame
        A table showing n_obj of all classes in columns, and 
        of all FOV and chips in rows.
    
    total_volume_obj_summary : pandas.DataFrame
        A table showing total_volume of all classes in columns, and 
        of all FOV and chips in rows.
    
    mean_intensity_obj_summary : pandas.DataFrame
        A table showing mean_intensity of all classes in columns, and 
        of all FOV and chips in rows.
    
    sum_intensity_obj_summary : pandas.DataFrame
        A table showing sum_intensity of all classes in columns, and 
        of all FOV and chips in rows.
        
    output_dir_path : string
        The path to the output directory
    """
    
    # Bar plots comparing features between classes in a FOV
    plot_feat_between_class(all_feat, output_dir_path)


    # Bar plots comparing features between FOV in a chip
    feature_name = "n_obj"
    plot_feat_between_fov(n_obj_summary, feature_name, output_dir_path)

    feature_name = "total_volume_obj"
    plot_feat_between_fov(total_volume_obj_summary, feature_name, output_dir_path)

    feature_name = "mean_intensity_obj"
    plot_feat_between_fov(mean_intensity_obj_summary, feature_name, output_dir_path)

    feature_name = "sum_intensity_obj"
    plot_feat_between_fov(sum_intensity_obj_summary, feature_name, output_dir_path)


    # Bar plots comparing features between chips
    feature_name = "n_obj"
    plot_feat_between_chip(n_obj_summary, feature_name, output_dir_path)

    feature_name = "total_volume_obj"
    plot_feat_between_chip(total_volume_obj_summary, feature_name, output_dir_path)

    feature_name = "mean_intensity_obj"
    plot_feat_between_chip(mean_intensity_obj_summary, feature_name, output_dir_path)

    feature_name = "sum_intensity_obj"
    plot_feat_between_chip(sum_intensity_obj_summary, feature_name, output_dir_path)


    # Stacked bar plots comparing features between FOV
    feature_name = "n_obj"
    stacked_bar_plot_between_fov(n_obj_summary, feature_name, output_dir_path)

    feature_name = "total_volume_obj"
    stacked_bar_plot_between_fov(total_volume_obj_summary, feature_name, output_dir_path)

    feature_name = "sum_intensity_obj"
    stacked_bar_plot_between_fov(sum_intensity_obj_summary, feature_name, output_dir_path)
    
    
    # Stacked bar plots comparing features between chips
    feature_name = "n_obj"
    stacked_bar_plot_between_chips(n_obj_summary, feature_name, output_dir_path)

    feature_name = "total_volume_obj"
    stacked_bar_plot_between_chips(total_volume_obj_summary, feature_name, output_dir_path)

    feature_name = "sum_intensity_obj"
    stacked_bar_plot_between_chips(sum_intensity_obj_summary, feature_name, output_dir_path)
    
    