"""
unit_tests.py

Description: 
    Unit test functions for checking correctness during development.
    
"""


__author__ = "John Lin"
__contact__ = "john.lin@emulatebio.com"
__date__ = "2023-05-28"
__status__ = "Prototype"


import numpy as np


# Unit tests
def check_id_uniqueness(subset_df_fov):
    """
    Check if chip id and fov id are unique
    
    Parameters
    ----------
    subset_df : numpy.DataFrame
        a subset of the full statistics table produced by the ImageJ
        script. Theoretically, it should contain only 1 unique Chip ID
        and 1 unique FOV ID.
        
    Raises
    ------
    AssertionError
        Unit test fails. chip id or fov id is not unique
    
    """
    
    # There are exactly 1 unique Chip ID and 1 unique FOV ID in the
    # data frame, no more or less
    assert len(subset_df_fov["Chip ID"].unique()) == 1
    assert len(subset_df_fov["FOV ID"].unique()) == 1

    #print("Test case passed")
    
    
def check_index_consistency(df, class_idx, class_pos):
    """
    Check if the indices after the points have been classified are
    consistent with the original full stats 

    Parameters
    ----------
    df : numpy.DataFrame
        subset_df_fov that is specific to the fov id
    class_idx : list of list
        contains indices that specifies which rows in the data frame
        belong to a particular class of objects (above, within, 
        bottom, below)
    class_pos: list of list
        contains x, y, z coordinates of all object centroids for each
        class

    Raises
    ------
    AssertionError
        Unit test fails. Indices do not match corresponding 
        x, y, z coordinates of object centroids.
        
    """
    
    # for every class of objects
    for i in range(len(class_pos)):
                
        class_i_idx = class_idx[i]
        class_i_pos = class_pos[i]
        
        class_i_pos = np.array(class_i_pos)
        class_i_idx = np.array(class_i_idx)
        #print("class_i_idx")
        #print(class_i_idx)
        #print("class_i_pos")
        #print(class_i_pos)


        # for every object in this class
        for i in range(len(class_i_idx)):
            
            # coordinates of objects accessed using index should be the same as the actual coordinates
            arr_from_full_stats = df.iloc[class_i_idx[i]][["X", "Y", "Z-pos (micron)"]].to_numpy()
            arr_from_classification = class_i_pos[i]

            #print(arr_from_full_stats)
            #print(arr_from_classification)
            #print(np.array_equal(arr_from_full_stats, arr_from_classification))
            assert np.array_equal(
                arr_from_full_stats, arr_from_classification
            )
        
    #print("Test case passed")
    
    
def check_pos_uniqueness(df, class_pos):
    """
    Check if x, y, and z coordinates of object centroids are unique

    Parameters
    ----------
    df : numpy.DataFrame
        subset_df_fov that is specific to the fov id
    class_pos: list of list
        contains x, y, z coordinates of all object centroids for each
        class
        
    Raises
    ------
    AssertionError
        Unit test fails. x, y, z coordinates of object centroids are
        not unique.
    
    """
    
    # for every class of objects
    for class_i_pos in class_pos:
        
        #print("class_i_pos")
        #print(class_i_pos)

        # for every object in this class
        for object_pos in class_i_pos:
            
            x = object_pos[0]
            y = object_pos[1]
            z = object_pos[2]
            #print(x)
            #print(y)
            #print(z)

            # There is exactly 1 matching coordinate in the data frame, no more or less
            #print(np.count_nonzero((df["X"] == x) & (df["Y"] == y) & (df["Z-pos (micron)"] == z)))
            assert 1 == np.count_nonzero(
                (df["X"] == x) & (df["Y"] == y) & (df["Z-pos (micron)"] == z)
            )

    #print("Test case passed")
    
    
def check_feat_consistency(features):
    """
    Check if sum of features match the overall features

    Parameters
    ----------
    features : numpy.DataFrame
        contains features for all classes, each in its row

    Raises
    ------
    AssertionError
        Unit test fails. Sum of features and overall features don't match.
    
    """
        
    #display(features)
    
    # The sum of n_obj for all classes should be equal to the overall n_obj
    sum_n_obj_all_classes = features["n_obj"].iloc[1:].sum(axis=0)
    n_obj_overall = features["n_obj"].iloc[0]
    
    #print(sum_n_obj_all_classes)
    #print(n_obj_overall)
    assert sum_n_obj_all_classes == n_obj_overall
    
    
    # The sum of volume for all classes should be equal to the overall volume
    sum_volume_all_classes = features["total_volume_obj"].iloc[1:].sum(axis=0)
    volume_overall = features["total_volume_obj"].iloc[0]
    
    #print(sum_volume_all_classes)
    #print(volume_overall)
    # Use np.isclose to check equality for floats
    assert np.isclose(
        sum_volume_all_classes, volume_overall,
        rtol=1e-05, atol=1e-08, equal_nan=False
    )
    
    
    # The mean intensity for all classes should be equal to the overall mean intensity 
    #mean_intensity_all_classes = (features["mean_intensity_obj"].iloc[1:] * features["n_obj"].iloc[1:]).sum(axis=0) / features["n_obj"].iloc[1:].sum(axis=0)
    #intensity_overall = features["mean_intensity_obj"].iloc[0]
    
    #print(mean_intensity_all_classes)
    #print(intensity_overall)
    # Use np.isclose to check equality for floats
    #assert np.isclose(
    #    mean_intensity_all_classes, intensity_overall,
    #    rtol=1e-05, atol=1e-08, equal_nan=False
    #)
    
    # The sum intensity for all classes should be equal to the overall sum intensity 
    sum_intensity_all_classes = features["sum_intensity_obj"].iloc[1:].sum(axis=0)
    sum_intensity_overall = features["sum_intensity_obj"].iloc[0]
    
    #print(sum_intensity_all_classes)
    #print(sum_intensity_overall)
    assert sum_intensity_all_classes == sum_intensity_overall

    
    #print("Test case passed")
    
    
    