"""
utils.py

Description: 
    Helper functions for file I/O.
    
"""


__author__ = "John Lin"
__contact__ = "john.lin@emulatebio.com"
__date__ = "2023-05-28"
__status__ = "Prototype"


import os
import datetime
import re

import pandas as pd
import matplotlib.pyplot as plt
from skimage import io
from gui import popup_message


def make_dir(dir_path):
    """
    Make a directory if doesn't exist.
    
    Parameters
    ----------
    dir_path : string
        The path to the directory
    """
    
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)
        
        
def csv_to_data_frame(dir_path, file_name):
    """
    A utility function that reads a Pandas data frame from a csv file
    
    Parameters
    ----------
    output_dir_path : string
        The path to the output directory
        
    file_name : string
        The filename of the csv file
    """
    file_path = os.path.join(dir_path, file_name)
    
    if os.path.exists(file_path):
        df = pd.read_csv(file_path)
        return df
    
    else:
        return None
    
    
def data_frame_to_csv(output_dir_path, output_category, file_name, df, df_index):
    """
    A utility function that writes a Pandas data frame to a csv file
    
    Parameters
    ----------
    output_dir_path : string
        The path to the output directory
             
    output_category : string
        The category of the output. Used for creating a sub-folder
        in the output directory.
        
    file_name : string
        The filename of the csv file
        
    df : pandas.DataFrame
        The data frame to write to file
        
    df_index : bool
        Write row names (index)
    """
    
    dir_path = os.path.join(output_dir_path, output_category)
    make_dir(dir_path)
    file_path = os.path.join(dir_path, file_name)
    df.to_csv(file_path, index=df_index)
    
    
def plot_to_png(output_dir_path, output_category, file_name, plt, bbox_inches=None):
    """
    A utility function that writes the matplotlib.pyplot object to a png file
    
    Parameters
    ----------
    output_dir_path : string
        The path to the output directory
        
    output_category : string
        The category of the output. Used for creating a sub-folder
        in the output directory.
        
    file_name : string
        The filename of the csv file
        
    plt : matplotlib.pyplot
        The plot to write to file
        
    bbox_inches : string
        Bounding box in inches: only the given portion of the figure is
        saved. If 'tight', try to figure out the tight bbox of the
        figure.
    """

    dir_path = os.path.join(output_dir_path, output_category)
    make_dir(dir_path)
    file_path = os.path.join(dir_path, file_name)
    plt.savefig(file_path, bbox_inches=bbox_inches)
    
    
def print_dash_line():
    """
    Print a dash line to separate different sections of tables
    """
    print("---------------------------------------------------------------------")
    
    
def img_info(img):
    """
    Print the shape of the image and the min and max pixel values
    
    Parameters
    ----------
    img : numpy.ndarray
        A numpy array of 2 to 4 dimensions
    """
    
    if len(img.shape) == 2:
        print("(H, W) =")
    elif len(img.shape) == 3:
        print("(N, H, W) =")
    elif len(img.shape) == 4:
        print("(N, H, W, C) =")

    print(img.shape)
    print(f"Max pixel value: {img.max()}")
    print(f"Min pixel value: {img.min()}")
    print("Image type = ")
    
    if len(img.shape) == 2:
        print(type(img[0][0]))
    elif len(img.shape) == 3:
        print(type(img[0][0][0]))
    elif len(img.shape) == 4:
        print(type(img[0][0][0][0]))
    
def display_img(img):
    """
    Display the image using plt.imshow
    
    Parameters
    ----------
    img : numpy.ndarray
        A 2D numpy array of an image
    """
    
    plt.imshow(img)
    plt.show()
    
    
def read_tiff(file_path):
    """
    Read the tiff image using io.imread.
    
    Parameters
    ----------       
    file_path : str
        Path to the image, with the extension.
        
    Returns
    -------
    img : numpy.ndarray
        A 3D image of object map in the shape of (N, H, W) or
        (N, H, W, C), where N is the number of z-slices. 
        
        If the image is an object map is produced by 3D
        object counter in ImageJ, the background is 0, 
        and objects are labeled by pixels with values equal to its
        index. All pixel values are unsigned int, which can be 8-bit,
        16-bit, or others, depending on the the output format set
        by ImageJ
        
    None
        If the specified file doesn't exist.
    """
    
    #file_path = os.path.join(dir_path, file_name)
    print(f"Reading {file_path}")
    
    # Check if the file exists
    if os.path.exists(file_path):

        img = io.imread(file_path, plugin="tifffile")
        img_info(img)
        return img
        
    else:
        print("The specified file doesn't exist.\n")
        return None
    
    
def write_tiff(output_dir_path, file_name, img, photometric=None):
    """
    A utility function that writes an object map to a TIFF file
    
    Parameters
    ----------
    output_dir_path : string
        The path to the output directory
        
    file_name : string
        The filename of the csv file
        
    img : numpy.ndarray
        A 3D image of object map in the shape of (N, H, W), where N is
        the number of z-slices.
        
    photometric : string
        The output format, passed to the tifffile plugin
    """
    
    make_dir(output_dir_path)
    file_path = os.path.join(output_dir_path, file_name)
    
    io.imsave(
        file_path, img, plugin="tifffile",
        photometric=photometric, check_contrast=False
    )
    
    
def write_logs(output_dir_path, logs):
    """
    Write logs to a txt file
    
    Parameters
    ----------
    output_dir_path : string
        The path to the output directory
        
    logs : list
        A list of strings, each of which is a log.
    """
    
    # Get the current time
    current_time = datetime.datetime.now()
    current_time = str(current_time)

    # Remove the microseconds
    current_time = current_time.split(".")[0]

    # Replace invalid char
    current_time = current_time.replace(" ", "_")
    current_time = current_time.replace(":", "-")

    #print(current_time)
    
    make_dir(output_dir_path)
    file_name = f'logs_{current_time}.txt'
    file_path = os.path.join(output_dir_path, file_name)
    
    # Write logs to file
    with open(file_path, 'w') as f:
        f.write('\n'.join(logs))
        
        
def _find_fov_id(file_path):
    """
    Use regular expression to get FOV ID from file path to the
    object map created by the ICR script in ImageJ. Called
    by find_fov_id().
    
    Sample format: "Objects map of r01c03f09ch1_compiled.tif"
    where r01c03f09 is the FOV ID, which must be preceded by
    "Objects map of " and ended in "ch"
    
    Parameters
    ----------
    file_path : string
        File path to the object map created by the ICR script in 
        ImageJ.
        
    Returns
    -------
    fov_id_match : list
        A list of strings that match the regular expression.
        The list is empty if there's no match.
    """
    
    fov_id_match = re.findall("(?<=Objects map of ).+(?=ch)", file_path)
    
    return fov_id_match


def find_fov_id(file_path_stain_A, file_path_stain_B):
    """
    Find the common FOV ID in file_path_stain_A and file_path_stain_B.
    
    If FOV ID is present and matches in file_path_stain_A
    and file_path_stain_B, return it as is.
    
    If no fov ID is found in either file name, or different in
    both file name, display an warning message and use the file
    name as the FOV ID instead.
    
    Parameters
    ----------
    file_path_stain_A : string
        The filename of the object map of stain A
        
    file_path_stain_B : string
        The filename of the object map of stain B
        
    Returns
    -------
    fov_id : string
        A string that represents the FOV ID. It can be the
        actual common FOV ID or just a string of both file names.
    """
    
    fov_id_stain_A = _find_fov_id(file_path_stain_A)
    fov_id_stain_B = _find_fov_id(file_path_stain_B)
    #print(fov_id_stain_A)
    #print(fov_id_stain_B)
    fov_id = ""

    # FOV ID is present and the equal in both file paths
    if fov_id_stain_A and fov_id_stain_A[0] == fov_id_stain_B[0]:
        fov_id = fov_id_stain_A[0]

    # No fov ID is found in either file name, or different in
    # both file name
    else:
        message = "Warning: FOV is not present or doesn't match in file names. Using file name as FOV ID instead. Program will continue."
        popup_message(message)
        print(message)

        # Get file name from path and remove extension
        fov_id_stain_A = file_path_stain_A.split("/")[-1].split(".")[0]
        fov_id_stain_B = file_path_stain_B.split("/")[-1].split(".")[0]

        fov_id = f"{fov_id_stain_A} and {fov_id_stain_B}"
        
    return fov_id


def get_tiff_list(dir_path):
    """
    Get the list of TIFF files in the input directory
    
    Parameters
    ----------
    dir_path : string
        Path to the input directory
        
    Returns
    -------
    file_list : list
        A list of strings, each of which is the file name
        of a TIFF file. The file name doesn't include the
        path.
    """
    
    file_list = []
    
    # Check if the directory exists and is not empty
    if os.path.isdir(dir_path) and os.listdir(dir_path):

        # Get the list of files
        temp_file_list = os.listdir(dir_path)


        # Keep only files with valid TIFF extension
        for file_name in temp_file_list:

            if ".tiff" in file_name or ".tif" in file_name:

                file_list.append(file_name)

    if not file_list:
        message = f"Warning: Directory is empty, contains no TIFF files, or doesn't exist. Nothing to do."
        print(message)
        popup_message(message)
        
    return file_list


def find_fov_id_tiff_stack(file_name):
    """
    Use regular expression to get FOV ID from file name to the
    raw z-stack image created by the ICR script in ImageJ. 
    
    Sample format: "r02c06f03ch3_compiled.tiff"
    where r02c06f03 is the FOV ID, which must be followed by "ch".
    Channel index is ch3, which starts with "ch" with only 1 digit,
    and ends with "_compiled".
    
    Parameters
    ----------
    file_name : string
        File name of the raw z-stack image created by the ICR script
        in ImageJ. The file name doens't include the path.
        
    Returns
    -------
    fov_id : string
        A string that represents the FOV ID. It can be the
        actual FOV ID or the file name if FOV ID is not present.
        
    ch_index : string
        A string that represents the channel index. It can be the
        actual channel index or an empty string if channel index
        is not present.
    """
    
    # Match strings that end with "ch"
    fov_id_match = re.findall(".+(?=ch)", file_name)
    
    # Match strings that start with "ch" with only 1 digit, and 
    # end with "_compiled"
    ch_index_match = re.findall("ch[0-9](?=_compiled)", file_name)
    
    #print(fov_id_match)
    #print(ch_index_match)

    # If FOV ID and channel index are present, return them.
    if fov_id_match and ch_index_match:
        
        fov_id = fov_id_match[0]
        ch_index = ch_index_match[0]
    
    # No fov ID or channel idnex are found, return the file
    # name instead
    else:
        
        # Remove extension
        fov_id = file_name.split(".")[0]
        ch_index = ""
    
    return fov_id, ch_index


